# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from dataclasses import dataclass
from datetime import datetime, date, timedelta, timezone
from decimal import Decimal
from typing import Any, Union

from ..damlast.builtins import builtins
from ..damlast import ExprVisitor, IdentityTypeVisitor
from ..damlast.daml_lf_1 import Block, BuiltinFunction, Case, Expr, PrimCon, PrimLit, Scenario, \
    Type, Update, ValName
from ..model.core import Party
from ..util.prim_types import frozendict


@dataclass(frozen=True)
class Constant:
    value: Any


class ExpressionEvaluator(ExprVisitor[Any], IdentityTypeVisitor):
    """
    Compute the result of an expression.
    """

    def visit_expr(self, expr: 'Union[Expr, Constant]'):
        if isinstance(expr, Constant):
            return expr.value
        else:
            return super(ExpressionEvaluator, self).visit_expr(expr)

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def visit_constant(self, constant: Constant) -> 'Any':
        return constant.value

    def visit_expr_var(self, var: str) -> 'Any':
        pass

    def visit_expr_val(self, val: 'ValName') -> 'Any':
        pass

    def visit_expr_builtin(self, builtin: 'BuiltinFunction') -> 'Any':
        return builtins.resolve(builtin).evaluate()

    def visit_expr_prim_con(self, prim_con: 'PrimCon') -> 'Any':
        if PrimCon.CON_TRUE == prim_con:
            return True
        elif PrimCon.CON_FALSE == prim_con:
            return False
        elif PrimCon.CON_UNIT == prim_con:
            return {}
        else:
            raise ValueError(f'unknown PrimCon value: {prim_con}')

    def visit_expr_prim_lit(self, prim_lit: 'PrimLit') -> 'Any':
        if prim_lit.int64 is not None:
            return prim_lit.int64
        elif prim_lit.decimal is not None:
            return Decimal(prim_lit.decimal)
        elif prim_lit.text is not None:
            return prim_lit.text
        elif prim_lit.timestamp is not None:
            return datetime.fromtimestamp(prim_lit.timestamp / 1e9, timezone.utc)
        elif prim_lit.party is not None:
            return Party(prim_lit.party)
        elif prim_lit.date is not None:
            return date(1970, 1, 1) + timedelta(days=prim_lit.date)
        else:
            raise ValueError(f'unknown PrimLit value: {prim_lit}')

    def visit_expr_rec_con(self, rec_con: 'Expr.RecCon') -> 'Any':
        return frozendict({fwt.field: self.visit_expr(fwt.expr) for fwt in rec_con.fields})

    def visit_expr_rec_proj(self, rec_proj: 'Expr.RecProj') -> 'Any':
        record_expr = self.visit_expr(rec_proj.record)
        return record_expr[rec_proj.field]

    def visit_expr_variant_con(self, variant_con: 'Expr.VariantCon') -> 'Any':
        ctor = variant_con.variant_con
        value = self.visit_expr(variant_con.variant_arg)
        return frozendict({ctor: value})

    def visit_expr_tuple_con(self, tuple_con: 'Expr.TupleCon') -> 'Any':
        return frozendict({fwt.field: self.visit_expr(fwt.expr) for fwt in tuple_con.fields})

    def visit_expr_tuple_proj(self, tuple_proj: 'Expr.TupleProj') -> 'Any':
        tuple_expr = self.visit_expr(tuple_proj.tuple)
        return tuple_expr[tuple_proj.field]

    def visit_expr_app(self, app: 'Expr.App') -> 'Any':
        fun = self.visit_expr(app.fun)
        args = [self.visit_expr(arg) for arg in app.args]
        return fun.apply(*args)

    def visit_expr_ty_app(self, ty_app: 'Expr.TyApp') -> 'Any':
        pass

    def visit_expr_abs(self, abs_: 'Expr.Abs') -> 'Any':
        pass

    def visit_expr_ty_abs(self, ty_abs: 'Expr.TyAbs') -> 'Any':
        pass

    def visit_expr_case(self, case: 'Case') -> 'Any':
        pass

    def visit_expr_let(self, let: 'Block') -> 'Any':
        current_scope = self
        for binding in let.bindings:
            var = binding.binder.var
            value = current_scope.eval_Expr(binding.bound)
            current_scope = current_scope.child_evaluator(value_bindings={var: Constant(value)})
        return current_scope.eval_Expr(let.body)

    def visit_expr_nil(self, nil: 'Expr.Nil') -> 'Any':
        pass

    def visit_expr_cons(self, cons: 'Expr.Cons') -> 'Any':
        pass

    def visit_expr_update(self, update: 'Update') -> 'Any':
        raise RuntimeError('expressions cannot currently be evaluated in the Update monad')

    def visit_expr_scenario(self, scenario: 'Scenario') -> 'Any':
        raise RuntimeError('expressions cannot currently be evaluated in the Scenario monad')

    def visit_expr_rec_upd(self, rec_upd: 'Expr.RecUpd') -> 'Any':
        raise RuntimeError('expressions cannot currently be evaluated in the Update monad')

    def visit_expr_tuple_upd(self, tuple_upd: 'Expr.TupleUpd') -> 'Any':
        raise RuntimeError('expressions cannot currently be evaluated in the Update monad')

    def visit_expr_none(self, none: 'Expr.None_') -> 'Any':
        return None

    def visit_expr_some(self, some: 'Expr.Some') -> 'Any':
        return self.visit_expr(some.body)

    def visit_type_var(self, var: 'Type.Var') -> 'Type':
        resolved_type = self.resolve_type(var.var)
        if resolved_type is not None:
            return self.visit_type(resolved_type)
        else:
            return Type(var=var)




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains a utility visitor for pretty-printing an expression tree.
"""

from io import StringIO
from typing import TextIO, NamedTuple, Union, Iterable, Optional

from ._rewriter import RewriteContext
from .expr_base import Expr, TRIGGER_ON_INIT
from .expr_impl import SuiteExpr, ProgramExpr, TemplateSelectsExpr, TemplateSelectExpr, \
    GetContractIdExpr, GetContractDataExpr, ConstantExpr, \
    TemplateFieldExpr, UpdateBlockExpr, CreateStatementExpr, ExerciseStatementExpr, Application, \
    AppExpr

from ..util.tools import boundary_iter

_PP = RewriteContext()


def pretty_print(expr: Expr) -> str:
    """
    Return a pretty-printed representation of an :class:`Expr`, optionally ANSI color-coded.
    """
    with StringIO() as buf:
        _PP.rewrite(expr, context=_PrettyPrintContext(buf, 0))
        return buf.getvalue()


class _PrettyPrintContext(NamedTuple):
    buf: TextIO
    indent: int

    def write(self, value: Union[None, str, Expr], **kwargs) -> None:
        if isinstance(value, str):
            self.buf.write(value)
        elif isinstance(value, Expr):
            _PP.rewrite(value, context=self, **kwargs)
        elif value is not None:
            raise ValueError(f'Only strings and Expr instances can be pretty printed (got {value})')

    def write_line(self, value: Union[None, str, Expr] = None) -> None:
        self.write(value)
        self.buf.write('\n')

    def write_keyword(self, text: str) -> None:
        self.buf.write(self.fmt.keyword(text))

    def write_keyword_line(self, text: str) -> None:
        self.buf.write(self.fmt.keyword(text))
        self.buf.write('\n')


####################################################################################################
# WRITER METHODS

@_PP.register(Application, recurse=False)
def _write_application(expr: Application, context: _PrettyPrintContext):
    for suite in expr.suites:
        context.write_line(suite)


@_PP.register(SuiteExpr, recurse=False)
def _write_suite(expr: SuiteExpr, context: _PrettyPrintContext):
    context.write_line(context.fmt.comments(f'-- rules for party {expr.party} --'))
    for program in expr.programs:
        context.write_line(program)


@_PP.register(ProgramExpr, recurse=False)
def _write_rule(expr: ProgramExpr, context: _PrettyPrintContext):
    context.write_keyword('rule ')
    context.write_line(f'{expr.name}')
    context.write(expr.select)
    if expr.trigger == TRIGGER_ON_INIT:
        context.write_keyword_line('    initially')
    else:
        context.write_keyword_line('    continually')
    context.write_keyword_line('    do')
    context.write(expr.update)


@_PP.register(TemplateSelectsExpr, recurse=False)
def _write_selects(expr: TemplateSelectsExpr, context: _PrettyPrintContext):
    context.write_keyword_line('  with')
    for tmpl in expr.select:
        context.write('    ')
        context.write(tmpl, include_type_info=True)
        context.write_line()

    context.write_keyword_line('  where')

    if expr.predicate is not None:
        context.write_keyword('    predicate ')
        context.write_line(expr.predicate)


@_PP.register(TemplateSelectExpr, recurse=False)
def _write_select(expr: TemplateSelectExpr, context: _PrettyPrintContext,
                  include_type_info: bool=False):
    if expr.alias is not None:
        context.write(expr.alias)
    else:
        context.write('_<' + expr.template.template_ref.name + '>')

    if include_type_info:
        context.write(' : ')
        context.write(expr.template.template_ref.name)


@_PP.register(AppExpr, recurse=False)
def _write_app(expr: AppExpr, context: _PrettyPrintContext):
    if expr.func.infix:
        _write_infix_operator(expr.func.name, expr.args, context)
    else:
        context.write(expr.func.name)
        for arg in expr.args:
            context.write(' ')
            context.write(arg)


@_PP.register(GetContractIdExpr, recurse=False)
def _write_cid(expr: GetContractIdExpr, context: _PrettyPrintContext):
    context.write('(')
    context.write_keyword('cid ')
    context.write(expr.expression)
    context.write(')')


@_PP.register(GetContractDataExpr, recurse=False)
def _write_cdata(expr: GetContractDataExpr, context: _PrettyPrintContext):
    context.write('(')
    context.write_keyword('cdata ')
    context.write(expr.expression)
    context.write(')')


@_PP.register(ConstantExpr, recurse=False)
def _write_constant(expr: ConstantExpr, context: _PrettyPrintContext):
    context.write(repr(expr.value))


@_PP.register(TemplateFieldExpr, recurse=False)
def _write_field(expr: TemplateFieldExpr, context: _PrettyPrintContext):
    context.write(expr.expression)
    context.write('.')
    context.write(expr.field_name)


@_PP.register(UpdateBlockExpr, recurse=False)
def _write_update(expr: UpdateBlockExpr, context: _PrettyPrintContext):
    if expr.statements:
        for stmt in expr.statements:
            context.write('      ')
            context.write_line(stmt)
    else:
        context.write_line('()')


@_PP.register(CreateStatementExpr, recurse=False)
def _write_create(expr: CreateStatementExpr, context: _PrettyPrintContext):
    context.write_keyword('create ')
    context.write(expr.template.name)
    _write_arguments(expr.arguments, context)


@_PP.register(ExerciseStatementExpr, recurse=False)
def _write_exercise(expr: ExerciseStatementExpr, context: _PrettyPrintContext):
    context.write_keyword('exercise ')
    context.write(expr.expression)
    context.write(' ')
    context.write(expr.choice_name)
    _write_arguments(expr.arguments, context)


####################################################################################################
# HELPER METHODS

def _write_infix_operator(infix_op: str,
                          expressions: Iterable[Expr],
                          context: _PrettyPrintContext) -> None:
    for is_last, child in boundary_iter(expressions):
        context.write(child)
        if not is_last:
            context.write(f' {infix_op} ')


def _write_arguments(arguments: Optional[dict], context: _PrettyPrintContext) -> None:
    if not arguments:
        return

    context.write_keyword(' with ')
    context.write('{')

    for is_last, (name, value) in boundary_iter(arguments.items()):
        context.write(name)
        context.write('=')
        if isinstance(value, Expr):
            context.write(value)
        else:
            context.write(repr(value))

        if not is_last:
            context.write(', ')

    context.write('}')



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains all of the magic for defining Expr and its subclasses.
"""

from typing import Dict, Any, Optional, NamedTuple, Set

from ..util.typing import unpack_optional

TRIGGER_ANY_TIME = 'TRIGGER_ANY_TIME'
TRIGGER_ON_INIT = 'TRIGGER_ON_INIT'

EQUALS = '='


class _Unquoted:
    __slots__ = ('value',)

    def __init__(self, value: str):
        self.value = value

    def __repr__(self):
        return self.value


def coerce_type(cls, obj):
    core_optional_type = unpack_optional(cls)
    if core_optional_type is not None:
        if obj is None:
            return None
        else:
            return coerce_type(core_optional_type, obj)

    if hasattr(cls, 'coerce'):
        return cls.coerce(obj)
    return obj


class FieldInfo(NamedTuple):
    name: str
    assignment: str
    type_annotation: Any
    type_imports: Set[Any]


def _field_assignment(field_name, type_annotation) -> FieldInfo:
    optional_type = unpack_optional(type_annotation)
    if optional_type is not None:
        assignment = f'self.{field_name} = coerce_type({optional_type.__name__}, {field_name}) ' \
                     f'if {field_name} is not None else None'
        return FieldInfo(
            name=field_name,
            assignment=assignment,
            type_annotation=type_annotation,
            type_imports={optional_type})
    elif hasattr(type_annotation, '__name__'):
        return FieldInfo(
            name=field_name,
            assignment=f'self.{field_name} = coerce_type({type_annotation.__name__}, {field_name})',
            type_annotation=type_annotation,
            type_imports={type_annotation})
    else:
        # TODO: Should this emit a warning?
        return FieldInfo(
            name=field_name,
            assignment=f'self.{field_name} = {field_name}',
            type_annotation=type_annotation,
            type_imports=set())


def _define_expr_data_class(name: str, annotated_fields: Dict[str, Any]) -> Optional[type]:
    """
    Create a simple type that defines some basic

    :return:
    """

    if not annotated_fields:
        return None

    field_infos = [_field_assignment(field_name, type_annotation)
                   for field_name, type_annotation in annotated_fields.items()]
    type_annotations = {im for fi in field_infos for im in fi.type_imports}

    data_class = f'_{name}Data'
    init_arg_list = ', '.join(annotated_fields.keys())

    init_assignments = '\n        '.join(fi.assignment for fi in field_infos)
    repr_fields = ', '.join(f'{field}={{self.{field}}}' for field in annotated_fields)
    newargs = repr(tuple(_Unquoted(f'self.{field}') for field in annotated_fields))

    ns = dict(coerce_type=coerce_type)
    ns.update({typ.__name__: typ for typ in type_annotations})
    exec_locals = {}

    class_def = f'''
class {data_class}:
    def __init__(self, {init_arg_list}):
        {init_assignments}

    def __repr__(self):
        return f'<{name}({repr_fields})>'

    def __getnewargs__(self):
        return {newargs}
'''
    try:
        exec(class_def, ns, exec_locals)
    except:
        print('Failed to construct an Expr representation from this string:')
        print(class_def)
        raise

    type_def = exec_locals[data_class]
    type_def.__annotations__ = annotated_fields
    return type_def


class _ExprMeta(type):
    """
    Metaclass for :class:`Expr` instances. Not meant to be used directly by outside callers.

    This expects a :class:`NamedTuple`-style syntax class definition, and does all the magic
    to actually turn it into an implementation. The generated class is compatible with rewriting.
    """
    def __new__(mcs, name, bases, dct):
        data_base_class = _define_expr_data_class(name, dct.get('__annotations__', {}))
        if data_base_class is not None:
            bases = (data_base_class,) + bases
        return super().__new__(mcs, name, bases, dct)


class Expr(metaclass=_ExprMeta):
    """
    An expression.
    """

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, cls):
            return obj
        else:
            raise TypeError(f"Couldn't convert to a {cls.__name__!r}: {obj!r}")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    ################################################################################################
    # COMPARISON FUNCTIONS

    def __lt__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of a less-than comparison of this expression with
        another expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_LT
        return AppExpr(BUILTIN_LT, (self, exprify_value(other)))

    def __le__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of a less-than-or-equal comparison of this
        expression with another expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_LE
        return AppExpr(BUILTIN_LE, (self, exprify_value(other)))

    def __eq__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of comparing this expression with another
        expression for equality.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_EQ
        return AppExpr(BUILTIN_EQ, (self, exprify_value(other)))

    def __ne__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of comparing this expression with another
        expression for inequality.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_NE
        return AppExpr(BUILTIN_NE, (self, exprify_value(other)))

    def __gt__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of a greater-than comparison of this expression with
        another expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_GT
        return AppExpr(BUILTIN_GT, (self, exprify_value(other)))

    def __ge__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of a greater-than-or-equal comparison of this
        expression with another expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_GE
        return AppExpr(BUILTIN_GE, (self, exprify_value(other)))

    ################################################################################################
    # MATHEMATICAL FUNCTIONS

    def __add__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of summing this expression with another expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_ADD
        return AppExpr(BUILTIN_ADD, (self, exprify_value(other)))

    def __sub__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of subtracting another expression from this
        expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_SUB
        return AppExpr(BUILTIN_SUB, (self, exprify_value(other)))

    def __mul__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of multiplying this expression by another
        expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_MUL
        return AppExpr(BUILTIN_MUL, (self, exprify_value(other)))

    def __truediv__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of dividing this expression by another
        expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_MUL
        return AppExpr(BUILTIN_MUL, (self, exprify_value(other)))

    def __concat__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of summing this expression with another expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_CONCAT
        return AppExpr(BUILTIN_CONCAT, (self, exprify_value(other)))

    def __pow__(self, other: 'Expr') -> 'Expr':
        """
        Return an expression that is the result of raising this expression by another expression.
        """
        from .expr_impl import AppExpr
        from .builtins import BUILTIN_POW
        return AppExpr(BUILTIN_POW, (self, exprify_value(other)))


def exprify_args(arguments):
    return {key: exprify_value(value) for key, value in arguments.items()}


def exprify_value(value):
    """
    Converts the specified unknown value to a subclass of `Expr` that is suitable to be used as an
    argument to a contract creation or choice exercising.
    """
    from .expr_impl import ConstantExpr, GetContractIdExpr, TemplateExpr, TemplateSelectExpr
    if isinstance(value, (TemplateExpr, TemplateSelectExpr)):
        # in an argument context, a bare template reference actually needs to resolve to the
        # contract ID of the default alias for a given template ID
        return GetContractIdExpr.coerce(value)

    elif isinstance(value, Expr):
        return value

    return ConstantExpr(value)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from .daml_lf_1 import *
from ..model.types import ModuleRef, TypeReference


# noinspection PyPep8Naming,PyMethodMayBeStatic
class ProtobufParser:
    def __init__(self, current_package: str):
        self.current_package = current_package

    # noinspection PyUnusedLocal
    def parse_Unit(self, pb) -> 'Unit':
        return UNIT

    def parse_ModuleRef(self, pb) -> 'Optional[ModuleRef]':
        sum_name = pb.package_ref.WhichOneof('Sum')
        if sum_name is None:
            return None
        elif sum_name == 'self':
            return ModuleRef(self.current_package, pb.module_name.segments)
        elif sum_name == 'package_id':
            return ModuleRef(pb.package_ref.package_id, pb.module_name.segments)
        else:
            raise ValueError(f'unknown sum type value: {sum_name!r}')

    def parse_TypeConName(self, pb) -> 'TypeReference':
        return TypeReference(
            self.parse_ModuleRef(pb.module),
            self.parse_DottedName(pb.name).segments)

    def parse_DottedName(self, pb) -> 'DottedName':
        return DottedName(segments=pb.segments)

    def parse_ValName(self, pb) -> 'ValName':
        return ValName(self.parse_ModuleRef(pb.module), tuple(pb.name))

    def parse_FieldWithType(self, pb) -> 'FieldWithType':
        """A field definition in a record or a variant associated with a type."""
        return FieldWithType(
            pb.field,
            self.parse_Type(pb.type))

    def parse_VarWithType(self, pb) -> 'VarWithType':
        """Binder associated with a type."""
        return VarWithType(
            pb.var,
            self.parse_Type(pb.type))

    def parse_TypeVarWithKind(self, pb) -> 'TypeVarWithKind':
        return TypeVarWithKind(
            pb.var,
            self.parse_Kind(pb.kind))

    def parse_FieldWithExpr(self, pb) -> 'FieldWithExpr':
        return FieldWithExpr(
            pb.field,
            self.parse_Expr(pb.expr))

    def parse_Binding(self, pb) -> 'Binding':
        return Binding(
            self.parse_VarWithType(pb.binder),
            self.parse_Expr(pb.bound))

    def parse_Kind(self, pb) -> 'Optional[Kind]':
        sum_name = pb.WhichOneof('Sum')
        if sum_name is None:
            return None
        elif sum_name == 'star':
            return Kind(star=self.parse_Unit(pb.star))
        elif sum_name == 'arrow':
            return Kind(arrow=self.parse_Kind_Arrow(pb.arrow))
        else:
            raise ValueError(f'unknown sum type value: {sum_name!r}')

    def parse_Kind_Arrow(self, pb) -> 'Kind.Arrow':
        return Kind.Arrow(
            tuple(self.parse_Kind(param) for param in pb.params),
            self.parse_Kind(pb.result))

    def parse_PrimType(self, pb) -> 'PrimType':
        return PrimType(pb)

    def parse_Type(self, pb) -> 'Type':
        sum_name = pb.WhichOneof('Sum')
        if sum_name == 'var':
            return self.parse_Type_Var(pb.var)
        elif sum_name == 'con':
            return self.parse_Type_Con(pb.con)
        elif sum_name == 'prim':
            return Type(prim=self.parse_Type_Prim(pb.prim))
        elif sum_name == 'fun':
            return self.parse_Type_Fun(pb.fun)
        elif sum_name == 'forall':
            return Type(forall=self.parse_Type_Forall(pb.forall))
        elif sum_name == 'tuple':
            return Type(tuple=self.parse_Type_Tuple(pb.tuple))
        else:
            raise ValueError(f'unknown sum type value: {sum_name!r}')

    def parse_Type_Var(self, pb) -> 'Type':
        return Type(var=Type.Var(pb.var, tuple(self.parse_Type(arg) for arg in pb.args)))

    def parse_Type_Con(self, pb, allow_rewrite: bool = True) -> 'Type':
        """
        Create a :class:`Type` instance (but may produce something slightly different than the AST
        due to ``Map``/``Optional`` type rewriting).
        """
        tycon = self.parse_TypeConName(pb.tycon)
        args = tuple(self.parse_Type(arg) for arg in pb.args)
        core_type = Type(con=Type.Con(tycon, args))

        if not allow_rewrite:
            return core_type
        if tycon.full_name_unambiguous == 'DA.Map:Map':
            return Type(prim=Type.Prim(PrimType.MAP_GENERIC, args, internal_type=core_type))
        elif tycon.full_name_unambiguous == 'DA.Internal.Prelude:Optional':
            return Type(prim=Type.Prim(PrimType.OPTIONAL, args, internal_type=core_type))
        else:
            return core_type

    def parse_Type_Prim(self, pb) -> 'Type.Prim':
        return Type.Prim(
            self.parse_PrimType(pb.prim),
            tuple(self.parse_Type(arg) for arg in pb.args))

    def parse_Type_Fun(self, pb) -> 'Type':
        """``fun`` has been deprecated and this code path is now replaced with PrimType.ARROW"""
        last_type = self.parse_Type(pb.result)
        for param in reversed(pb.params):
            last_type = Type(prim=Type.Prim(PrimType.ARROW, (self.parse_Type(param), last_type)))
        return last_type

    def parse_Type_Forall(self, pb) -> 'Type.Forall':
        return Type.Forall(
            tuple(self.parse_TypeVarWithKind(var) for var in pb.vars),
            self.parse_Type(pb.body))

    def parse_Type_Tuple(self, pb) -> 'Type.Tuple':
        return Type.Tuple(tuple(self.parse_FieldWithType(field) for field in pb.fields))

    def parse_PrimCon(self, pb) -> 'Optional[PrimCon]':
        if pb is None:
            return None
        elif pb == 0:
            return PrimCon.CON_UNIT
        elif pb == 1:
            return PrimCon.CON_FALSE
        elif pb == 2:
            return PrimCon.CON_TRUE
        else:
            raise ValueError(f'unknown enum value: {pb!r}')

    def parse_BuiltinFunction(self, pb) -> 'Optional[BuiltinFunction]':
        if pb is None:
            return None
        return BuiltinFunction(pb)

    def parse_PrimLit(self, pb) -> 'PrimLit':
        sum_name = pb.WhichOneof('Sum')
        if sum_name == 'int64':
            return PrimLit(int64=pb.int64)
        elif sum_name == 'decimal':
            return PrimLit(decimal=pb.decimal)
        elif sum_name == 'text':
            return PrimLit(text=pb.text)
        elif sum_name == 'timestamp':
            return PrimLit(timestamp=pb.timestamp)
        elif sum_name == 'party':
            return PrimLit(party=pb.party)
        elif sum_name == 'date':
            return PrimLit(date=pb.date)
        else:
            raise ValueError(f'unknown Sum value: {pb!r}')

    def parse_Expr(self, pb, optional: bool = False) -> 'Expr':
        location = self.parse_Location(pb.location)
        args = {'location': location}

        sum_name = pb.WhichOneof('Sum')
        if optional and sum_name is None:
            return None
        if sum_name == 'var':
            args['var'] = pb.var
        elif sum_name == 'val':
            args['val'] = self.parse_ValName(pb.val)
        elif sum_name == 'builtin':
            args['builtin'] = self.parse_BuiltinFunction(pb.builtin)
        elif sum_name == 'prim_con':
            args['prim_con'] = self.parse_PrimCon(pb.prim_con)
        elif sum_name == 'prim_lit':
            args['prim_lit'] = self.parse_PrimLit(pb.prim_lit)
        elif sum_name == 'rec_con':
            args['rec_con'] = self.parse_Expr_RecCon(pb.rec_con)
        elif sum_name == 'rec_proj':
            args['rec_proj'] = self.parse_Expr_RecProj(pb.rec_proj)
        elif sum_name == 'variant_con':
            args['variant_con'] = self.parse_Expr_VariantCon(pb.variant_con)
        elif sum_name == 'tuple_con':
            args['tuple_con'] = self.parse_Expr_TupleCon(pb.tuple_con)
        elif sum_name == 'tuple_proj':
            args['tuple_proj'] = self.parse_Expr_TupleProj(pb.tuple_proj)
        elif sum_name == 'app':
            args['app'] = self.parse_Expr_App(pb.app)
        elif sum_name == 'ty_app':
            args['ty_app'] = self.parse_Expr_TyApp(pb.ty_app)
        elif sum_name == 'abs':
            args['abs'] = self.parse_Expr_Abs(pb.abs)
        elif sum_name == 'ty_abs':
            args['ty_abs'] = self.parse_Expr_TyAbs(pb.ty_abs)
        elif sum_name == 'case':
            args['case'] = self.parse_Case(pb.case)
        elif sum_name == 'let':
            args['let'] = self.parse_Block(pb.let)
        elif sum_name == 'nil':
            args['nil'] = self.parse_Expr_Nil(pb.nil)
        elif sum_name == 'cons':
            args['cons'] = self.parse_Expr_Cons(pb.cons)
        elif sum_name == 'update':
            args['update'] = self.parse_Update(pb.update)
        elif sum_name == 'scenario':
            args['scenario'] = self.parse_Scenario(pb.scenario)
        elif sum_name == 'rec_upd':
            args['rec_upd'] = self.parse_Expr_RecUpd(pb.rec_upd)
        elif sum_name == 'tuple_upd':
            args['tuple_upd'] = self.parse_Expr_TupleUpd(pb.tuple_upd)
        elif sum_name == 'none':
            args['none'] = self.parse_Expr_None(pb.none)
        elif sum_name == 'some':
            args['some'] = self.parse_Expr_Some(pb.some)
        else:
            raise ValueError(f'Unknown type of Expr: {sum_name!r}')

        return Expr(**args)

    def parse_Expr_RecCon(self, pb) -> 'Expr.RecCon':
        return Expr.RecCon(
            self.parse_Type_Con(pb.tycon, allow_rewrite=False).con,
            tuple(self.parse_FieldWithExpr(field) for field in pb.fields))  # length > 0

    def parse_Expr_RecProj(self, pb) -> 'Expr.RecProj':
        return Expr.RecProj(
            self.parse_Type_Con(pb.tycon, allow_rewrite=False).con,  # Always fully applied
            pb.field,
            self.parse_Expr(pb.record))

    def parse_Expr_RecUpd(self, pb) -> 'Expr.RecUpd':
        """Set ``field`` in ``record`` to ``update``."""
        return Expr.RecUpd(
            self.parse_Type_Con(pb.tycon, allow_rewrite=False).con,
            pb.field,
            self.parse_Expr(pb.record),
            self.parse_Expr(pb.update))

    def parse_Expr_VariantCon(self, pb) -> 'Expr.VariantCon':
        return Expr.VariantCon(
            self.parse_Type_Con(pb.tycon, allow_rewrite=False).con,  # Always fully applied
            pb.variant_con,
            self.parse_Expr(pb.variant_arg))

    def parse_Expr_TupleCon(self, pb) -> 'Expr.TupleCon':
        return Expr.TupleCon(
            tuple(self.parse_FieldWithExpr(field) for field in pb.fields))  # length > 0

    def parse_Expr_TupleProj(self, pb) -> 'Expr.TupleProj':
        return Expr.TupleProj(
            pb.field,
            self.parse_Expr(pb.tuple))

    def parse_Expr_TupleUpd(self, pb) -> 'Expr.TupleUpd':
        """Set ``field`` in ``tuple`` to ``update``."""
        return Expr.TupleUpd(
            pb.field,
            self.parse_Expr(pb.tuple),
            self.parse_Expr(pb.update))

    def parse_Expr_App(self, pb) -> 'Expr.App':
        return Expr.App(
            self.parse_Expr(pb.fun),
            tuple(self.parse_Expr(arg) for arg in pb.args))  # length > 0

    def parse_Expr_TyApp(self, pb) -> 'Expr.TyApp':
        return Expr.TyApp(
            self.parse_Expr(pb.expr),
            tuple(self.parse_Type(type) for type in pb.types))  # length > 0

    def parse_Expr_Abs(self, pb):
        return Expr.Abs(
            tuple(self.parse_VarWithType(param) for param in pb.param),  # length > 0
            self.parse_Expr(pb.body))

    def parse_Expr_TyAbs(self, pb):
        return Expr.TyAbs(
            tuple(self.parse_TypeVarWithKind(param) for param in pb.param),  # length > 0
            self.parse_Expr(pb.body))

    def parse_Expr_Nil(self, pb):
        return Expr.Nil(
            self.parse_Type(pb.type))

    def parse_Expr_Cons(self, pb):
        return Expr.Cons(
            self.parse_Type(pb.type),
            tuple(self.parse_Expr(front) for front in pb.front),  # length > 0
            self.parse_Expr(pb.tail))

    def parse_Expr_None(self, pb):
        return Expr.None_(
            self.parse_Type(pb.type))

    def parse_Expr_Some(self, pb):
        return Expr.Some(
            self.parse_Type(pb.type),
            self.parse_Expr(pb.body))

    def parse_CaseAlt(self, pb) -> 'CaseAlt':
        body = self.parse_Expr(pb.body)
        sum_name = pb.WhichOneof('Sum')
        if sum_name == 'default':
            return CaseAlt(default=self.parse_Unit(pb.default), body=body)
        elif sum_name == 'variant':
            return CaseAlt(variant=self.parse_CaseAlt_Variant(pb.variant), body=body)
        elif sum_name == 'prim_con':
            return CaseAlt(prim_con=self.parse_PrimCon(pb.prim_con), body=body)
        elif sum_name == 'nil':
            return CaseAlt(nil=self.parse_Unit(pb.nil), body=body)
        elif sum_name == 'cons':
            return CaseAlt(cons=self.parse_CaseAlt_Cons(pb.cons), body=body)
        elif sum_name == 'none':
            return CaseAlt(none=self.parse_Unit(pb.none), body=body)
        elif sum_name == 'some':
            return CaseAlt(some=self.parse_CaseAlt_Some(pb.some), body=body)
        else:
            raise ValueError('unknown Sum value')

    def parse_CaseAlt_Variant(self, pb) -> 'CaseAlt.Variant':
        return CaseAlt.Variant(self.parse_TypeConName(pb.con), pb.variant, pb.binder)

    def parse_CaseAlt_Cons(self, pb) -> 'CaseAlt.Cons':
        return CaseAlt.Cons(pb.var_head, pb.var_tail)

    def parse_CaseAlt_Some(self, pb) -> 'CaseAlt.Some':
        return CaseAlt.Some(pb.var_body)

    def parse_Case(self, pb) -> 'Case':
        return Case(
            self.parse_Expr(pb.scrut),
            tuple(self.parse_CaseAlt(alt) for alt in pb.alts))

    def parse_Block(self, pb) -> 'Block':
        return Block(
            tuple(self.parse_Binding(binding) for binding in pb.bindings),
            self.parse_Expr(pb.body))

    def parse_Pure(self, pb) -> 'Pure':
        return Pure(
            type=self.parse_Type(pb.type),
            expr=self.parse_Expr(pb.expr))

    def parse_Update(self, pb) -> 'Update':
        sum_name = pb.WhichOneof('Sum')
        if sum_name == 'pure':
            return Update(pure=self.parse_Pure(pb.pure))
        elif sum_name == 'block':
            return Update(block=self.parse_Block(pb.block))
        elif sum_name == 'create':
            return Update(create=self.parse_Update_Create(pb.create))
        elif sum_name == 'exercise':
            return Update(exercise=self.parse_Update_Exercise(pb.exercise))
        elif sum_name == 'fetch':
            return Update(fetch=self.parse_Update_Fetch(pb.fetch))
        elif sum_name == 'get_time':
            return Update(get_time=self.parse_Unit(pb.get_time))
        elif sum_name == 'embed_expr':
            return Update(embed_expr=self.parse_Update_EmbedExpr(pb.embed_expr))
        elif sum_name == 'lookup_by_key':
            return Update(embed_expr=self.parse_Update_RetrieveByKey(pb.lookup_by_key))
        elif sum_name == 'fetch_by_key':
            return Update(embed_expr=self.parse_Update_RetrieveByKey(pb.fetch_by_key))
        else:
            raise ValueError(f'unknown Sum value: {sum_name!r}')

    def parse_Update_Create(self, pb) -> 'Update.Create':
        return Update.Create(
            template=self.parse_TypeConName(pb.template),
            expr=self.parse_Expr(pb.expr))

    def parse_Update_Exercise(self, pb) -> 'Update.Exercise':
        return Update.Exercise(
            template=self.parse_TypeConName(pb.template),
            choice=pb.choice,
            cid=self.parse_Expr(pb.cid),
            actor=self.parse_Expr(pb.actor, optional=True),
            arg=self.parse_Expr(pb.arg))

    def parse_Update_Fetch(self, pb) -> 'Update.Fetch':
        return Update.Fetch(
            template=self.parse_TypeConName(pb.template),
            cid=self.parse_Expr(pb.cid))

    def parse_Update_EmbedExpr(self, pb) -> 'Update.EmbedExpr':
        return Update.EmbedExpr(
            type=self.parse_Type(pb.type),
            body=self.parse_Expr(pb.body))

    def parse_Update_RetrieveByKey(self, pb) -> 'Update.RetrieveByKey':
        return Update.RetrieveByKey(
            template=self.parse_TypeConName(pb.template),
            key=self.parse_Expr(pb.key))

    def parse_Scenario(self, pb) -> 'Scenario':
        sum_name = pb.WhichOneof('Sum')
        if sum_name == 'pure':
            return Scenario(pure=self.parse_Pure(pb.pure))
        elif sum_name == 'block':
            return Scenario(block=self.parse_Block(pb.block))
        elif sum_name == 'commit':
            return Scenario(commit=self.parse_Scenario_Commit(pb.commit))
        elif sum_name == 'mustFailAt':
            return Scenario(must_fail_at=self.parse_Scenario_Commit(pb.mustFailAt))
        elif sum_name == 'pass':
            return Scenario(must_fail_at=self.parse_Expr(getattr(pb, 'pass')))
        elif sum_name == 'get_time':
            return Scenario(get_time=self.parse_Unit(pb.get_time))
        elif sum_name == 'get_party':
            return Scenario(get_party=self.parse_Expr(pb.get_party))
        elif sum_name == 'embed_expr':
            return Scenario(embed_expr=self.parse_Scenario_EmbedExpr(pb.embed_expr))
        else:
            raise ValueError('unknown Sum value')

    def parse_Scenario_Commit(self, pb) -> 'Scenario.Commit':
        return Scenario.Commit(
            party=self.parse_Expr(pb.party),
            expr=self.parse_Expr(pb.expr),
            ret_type=self.parse_Type(pb.ret_type))

    def parse_Scenario_EmbedExpr(self, pb) -> 'Scenario.EmbedExpr':
        return Scenario.EmbedExpr(
            type=self.parse_Type(pb.type),
            body=self.parse_Expr(pb.body))

    def parse_Location(self, pb) -> 'Location':
        return Location(
            self.parse_ModuleRef(pb.module),
            self.parse_Location_Range(pb.range))

    def parse_Location_Range(self, pb) -> 'Location.Range':
        return Location.Range(
            pb.start_line,
            pb.start_col,
            pb.end_line,
            pb.end_col)

    def parse_TemplateChoice(self, pb) -> 'TemplateChoice':
        return TemplateChoice(
            name=pb.name,
            consuming=pb.consuming,
            controllers=self.parse_Expr(pb.controllers),
            arg_binder=self.parse_VarWithType(pb.arg_binder),
            ret_type=self.parse_Type(pb.ret_type),
            update=self.parse_Expr(pb.update),
            self_binder=pb.self_binder,
            location=self.parse_Location(pb.location))

    def parse_DefTemplate(self, pb) -> 'DefTemplate':
        return DefTemplate(
            tycon=self.parse_DottedName(pb.tycon),
            param=pb.param,
            precond=self.parse_Expr(pb.precond),
            signatories=self.parse_Expr(pb.signatories),
            agreement=self.parse_Expr(pb.agreement),
            choices=tuple(self.parse_TemplateChoice(choice) for choice in pb.choices),
            observers=self.parse_Expr(pb.observers),
            location=self.parse_Location(pb.location))

    def parse_DefDataType(self, pb) -> 'DefDataType':
        kwargs = dict(
            name=self.parse_DottedName(pb.name),
            params=tuple(self.parse_TypeVarWithKind(param) for param in pb.params),
            serializable=pb.serializable,
            location=self.parse_Location(pb.location))
        DataCons_name = pb.WhichOneof('DataCons')
        if DataCons_name == 'record':
            kwargs['record'] = self.parse_DefDataType_Fields(pb.record)
        elif DataCons_name == 'variant':
            kwargs['variant'] = self.parse_DefDataType_Fields(pb.variant)
        return DefDataType(**kwargs)

    def parse_DefDataType_Fields(self, pb) -> 'DefDataType.Fields':
        return DefDataType.Fields(
            fields=tuple(self.parse_FieldWithType(field) for field in pb.fields))

    def parse_DefValue(self, pb) -> 'DefValue':
        return DefValue(
            name_with_type=self.parse_DefValue_NameWithType(pb.name_with_type),
            expr=self.parse_Expr(pb.expr),
            no_party_literals=pb.no_party_literals,
            is_test=pb.is_test,
            location=self.parse_Location(pb.location))

    def parse_DefValue_NameWithType(self, pb) -> 'DefValue.NameWithType':
        return DefValue.NameWithType(
            name=tuple(pb.name),
            type=self.parse_Type(pb.type))

    def parse_FeatureFlags(self, pb) -> 'FeatureFlags':
        return FeatureFlags(
          forbidPartyLiterals=pb.forbidPartyLiterals,
          dontDivulgeContractIdsInCreateArguments=pb.dontDivulgeContractIdsInCreateArguments,
          dontDiscloseNonConsumingChoicesToObservers=pb.dontDiscloseNonConsumingChoicesToObservers)

    def parse_Module(self, pb) -> 'Module':
        return Module(
            name=self.parse_DottedName(pb.name),
            flags=self.parse_FeatureFlags(pb.flags),
            data_types=tuple(self.parse_DefDataType(data_type) for data_type in pb.data_types),
            values=tuple(self.parse_DefValue(value) for value in pb.values),
            templates=tuple(self.parse_DefTemplate(template) for template in pb.templates))

    def parse_Package(self, pb) -> 'Package':
        return Package(modules=tuple(self.parse_Module(module) for module in pb.modules))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Simple command-line handlers.
"""

import logging
import sys
from typing import List

from .. import setup_default_logger
from ..model.core import ConfigurationError
from ._base import CliCommand
from .fetch import FetchComponentCommand
from .ls import ListAllCommand
from .metadata import PrintMetadataCommand
from .package import PackageDarCommand
from .sandbox import SandboxCommand
from .tail import TailCommand
from .version import VersionCommand

COMMANDS = [
    FetchComponentCommand(),
    ListAllCommand(),
    PrintMetadataCommand(),
    PackageDarCommand(),
    SandboxCommand(),
    TailCommand(),
    VersionCommand(),
]  # type: List[CliCommand]


def main():
    """
    Executes one of the known commands.
    """
    if len(sys.argv) > 1:
        command = sys.argv[1]
        command_args = sys.argv[2:]

        for cmd in COMMANDS:
            if cmd.name == command:
                try:
                    sys.exit(run(cmd, command_args))
                except ConfigurationError as error:
                    for reason in error.reasons:
                        print(reason)
                    sys.exit(-1)

        print("Unknown command: " + command)

    print_cmd_help()


def run(cmd, args) -> int:
    parser = cmd.parser()
    parsed_args = parser.parse_args(args)

    log_level = getattr(parsed_args, 'log_level', logging.WARNING)
    if log_level is None:
        log_level = logging.WARNING

    setup_default_logger(level=log_level)
    return cmd.execute(parsed_args)


def print_cmd_help():
    print("Valid commands:")
    for cmd in COMMANDS:
        if not cmd.hidden:
            print(cmd.name)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains helper methods for detecting terminal capabilities.
"""

from subprocess import Popen, DEVNULL, PIPE
from typing import Optional, Tuple


def termsize() -> Tuple[Optional[int], Optional[int]]:
    """
    Return the current size of the terminal. If the current terminal is not a tty, then
    ``(None, None)`` is returned.
    """
    with Popen(['stty', 'size'], stdout=PIPE, stderr=DEVNULL) as proc:
        term_size_str = proc.stdout.read().decode('utf8')
    if term_size_str is not None:
        dimensions = term_size_str.split()
        if len(dimensions) >= 2:
            return int(dimensions[0]), int(dimensions[1])
        else:
            return None, None
    else:
        return None, None


def print_termcap() -> None:
    """
    Output discovered settings to the terminal.
    """
    rows, cols = termsize()
    if rows is not None and cols is not None:
        print('The terminal size is:')
        print(termsize())
    else:
        print('Not a terminal.')


if __name__ == '__main__':
    print_termcap()



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from dataclasses import dataclass
from typing import Optional, Sequence

from .. import LOG
from .fetch import sdk_component_path
from ..util.process import ProcessContext


@dataclass
class PackageOptions:
    """
    Options that control how DAML packages are constructed.
    """
    files: 'Sequence[str]'
    output_path: str
    extra_args: 'Sequence[str]'


def package(options: 'PackageOptions', component: 'Optional[str]' = None) -> 'ProcessContext':
    name, path = sdk_component_path(component or 'damlc')

    if name != 'damlc':
        raise ValueError(f'Unknown packaging component: {component}')

    LOG.info('Using %s to create a DAR package...', name)
    args = [path / 'da-hs-damlc-app', *_damlc_package_options(options)]
    return ProcessContext(args, logger=logging.getLogger('damlc'))


def _damlc_package_options(options: 'PackageOptions') -> 'Sequence[str]':
    if options.extra_args:
        return ['package', *options.files, *options.extra_args, 'package-name', '-o', options.output_path]
    else:
        return ['package', *options.files, 'package-name', '-o', options.output_path]





# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import StringIO
from typing import Mapping, Optional, Sequence, Union

from ._render_base import PrettyPrintBase, pretty_print_syntax
from .util import maybe_parentheses, indent
from .. import LOG
from ..damlast.daml_lf_1 import DefDataType, DefTemplate, Expr, Type, Case, Block, \
    Update, Scenario, PrimType
from ..damlast.util import def_value, list_type, PARTY_TYPE
from ..model.types import ModuleRef, Type as OldType, TypeReference
from ..model.types_store import PackageStore


def values_by_module(store: 'PackageStore') \
        -> 'Mapping[ModuleRef, Mapping[Sequence[str], Union[Expr, OldType]]]':
    from collections import defaultdict
    d = defaultdict(defaultdict)
    for vn, vv in store._value_types.items():
        d[vn.module][vn.name] = vv
    for vn, vv in store._data_types.items():
        d[vn.module][vn.name] = vv
    return d


@pretty_print_syntax('python')
class PythonPrettyPrint(PrettyPrintBase):
    """
    Renderer for :class:`Expr` and :class:`Type` hierarchies that output text that look like
    Python code.
    """

    syntax = 'python'

    def lexer(self):
        try:
            from pygments.lexers.python import PythonLexer
        except ImportError:
            return None
        return PythonLexer()

    def visit_module_ref_start(self, module_ref: 'ModuleRef') -> str:
        return f'@module\nclass {module_ref.module_name[-1]}:\n' if module_ref.module_name else ''

    # def visit_def_value(self, def_value: 'DefValue', as_instance_method: bool = False):
    #     def_value.name_with_type

    def visit_empty_block_body(self):
        return 'pass'

    def visit_def_template(
            self,
            template: 'Optional[DefTemplate]',
            def_data_type: 'Optional[DefDataType]' = None) -> str:
        template_full_name = TypeReference(self.context.current_module, def_data_type.name.segments).full_name
        template_name = def_data_type.name.segments[-1]
        slot_names = tuple(fwt.field for fwt in def_data_type.record.fields) if def_data_type.record else ()
        signatories_def = def_value('signatories', list_type(PARTY_TYPE), template.signatories) if template is not None else None

        if template is not None:
            lines = []
            lines.append(f'class {template_name}(metaclass=TemplateMeta, template_name={template_full_name!r}):')
            lines.append('    """')
            lines.append('    Example usage:')
            lines.append(f'        create({template_full_name},')
            lines.append(f'               {python_example_object(self.store, def_data_type)})')
            for choice in template.choices:
                lines.append(f'        exercise(cid, {choice.name!r}, {python_example_object(self.store, choice.arg_binder.type)})')
            lines.append('    """')
            lines.append(indent(self._visit_def_type_body(slot_names), 4))
            for choice in template.choices:
                ct = choice.arg_binder.type
                if ct.prim is not None and ct.prim.prim == PrimType.UNIT:
                    choice_slot_names = ()
                else:
                    tt = self.store.resolve_type_reference(ct.con.tycon)
                    choice_slot_names = tuple(name for name, _ in tt.named_args) if tt is not None else ()
                lines.append(f'    class {choice.name}(metaclass=ChoiceMeta, template_name={template_full_name!r}, choice_name={choice.name!r}):')
                lines.append(indent(self._visit_def_type_body(choice_slot_names), 8))
            return '\n'.join(lines)
        else:
            return ''

    def _visit_def_type_body(self, slot_names: 'Sequence[str]') -> str:
        lines = [f'__slots__ = {slot_names!r}\n']
        if slot_names:
            lines.append(f'def __init__(self, {", ".join(slot_names)}):')
            lines.extend(f'    self.{slot_name} = {slot_name}' for slot_name in slot_names)
            lines.append('')
            lines.append(f'def _asdict(self) -> dict:')
            lines.append(f'    return {{')
            lines.extend(f'        {slot_name!r}: self.{slot_name},' for slot_name in slot_names)
            lines.append(f'        }}')
        else:
            lines.append('def _asdict(self) -> dict:')
            lines.append('    return {}')
        return '\n'.join(lines) + '\n'

    def visit_expr_prim_con_inline_unit(self):
        return '{}'

    def visit_expr_rec_con_inline(self, rec_con: 'Expr.RecCon'):
        with StringIO() as buf:
            buf.write('record(')
            delim = ''
            for fwe in rec_con.fields:
                buf.write(delim)
                buf.write(fwe.field)
                buf.write('=')
                expr = self.visit_expr(fwe.expr)
                if ' ' in expr:
                    buf.write('(')
                buf.write(expr)
                if ' ' in expr:
                    buf.write(')')
                delim = ', '
            buf.write(')')

            # if self.decl_name is not None:
            #     buf.write('  # type: ')
            #     buf.write(self.visit_type_con(rec_con.tycon))
            return buf.getvalue()

    def visit_expr_rec_proj(self, rec_proj: 'Expr.RecProj') -> str:
        record_text = self.visit_expr(rec_proj.record)
        if ' ' in record_text:
            return f'({record_text})[{rec_proj.field!r}]'
        else:
            return f'{record_text}[{rec_proj.field!r}]'

    def visit_expr_variant_con(self, variant_con: 'Expr.VariantCon') -> str:
        with StringIO() as buf:
            try:
                scope = self
                if self.context.decl_name is not None:
                    buf.write('.'.join(self.context.decl_name))
                    buf.write(' = ')
                else:
                    buf.write('return ')
                buf.write('variant(')
                buf.write(variant_con.variant_con)
                buf.write('=')
                expr = scope.visit_expr(variant_con.variant_arg)
                if ' ' in expr:
                    buf.write('(')
                buf.write(expr)
                if ' ' in expr:
                    buf.write(')')
                buf.write(')')

                if self.context.decl_name is not None:
                    buf.write('  # type: ')
                    buf.write(self.visit_type_con(variant_con.tycon))
                return buf.getvalue()
            except:
                LOG.exception('why why why')
                exit(-1)

    def visit_expr_tuple_con(self, tuple_con: 'Expr.TupleCon') -> str:
        if tuple_con.fields:
            with StringIO() as buf:
                delim = 'tuple('
                for fwe in tuple_con.fields:
                    buf.write(delim)
                    buf.write(fwe.field)
                    buf.write('=')
                    expr = self.visit_expr(fwe.expr)
                    if ' ' in expr:
                        buf.write('(')
                    buf.write(expr)
                    if ' ' in expr:
                        buf.write(')')
                    delim = ', '
                buf.write(')')
                return buf.getvalue()
        else:
            return 'tuple()'

    def visit_expr_tuple_proj(self, tuple_proj: 'Expr.TupleProj') -> str:
        record_text = self.visit_expr(tuple_proj.tuple)
        if ' ' in record_text:
            return f'({record_text})[{tuple_proj.field!r}]'
        else:
            return f'{record_text}[{tuple_proj.field!r}]'

    def visit_expr_app_inline(self, app: 'Expr.App'):
        fun_text = maybe_parentheses(self.visit_expr(app.fun))
        args_text = ', '.join(self.visit_expr(arg) for arg in app.args)
        return f'{fun_text}({args_text})'

    def visit_expr_ty_app(self, ty_app: 'Expr.TyApp') -> str:
        fun_text = self.visit_expr(ty_app.expr)
        return fun_text
        args_text = ', '.join(self.visit_type(arg) for arg in ty_app.types)
        return f'ty_apply({fun_text})<{args_text}>'

    def visit_expr_abs_decl(self, abs_: 'Expr.Abs', context: 'CodeContext'):
        body = self.visit_expr(abs_.body)
        try:
            body_type = self.type_computer.visit_expr(abs_.body)
            body_type_expr = ' -> \'' + self.visit_type(body_type) + '\''
        except:
            # LOG.exception("Couldn't compute a type!")
            body_type_expr = ''
        safe_decl_name = '.'.join(context.decl_name).replace('$', '_')
        args = ', '.join(f'{p.var}: \'{self.visit_type(p.type)}\'' for p in abs_.param)
        decl_line = f'@staticmethod\ndef {safe_decl_name}({args}){body_type_expr}:\n'
        return decl_line + ''.join(('    ' + line) for line in body.splitlines(keepends=True))

    def visit_expr_abs_inline(self, abs_: 'Expr.Abs'):
        body = self.visit_expr(abs_.body)
        args = ', '.join(f'{p.var}' for p in abs_.param)
        if args:
            args = ' ' + args
        return f'lambda{args}: {body}'

    def visit_expr_ty_abs(self, ty_abs: 'Expr.TyAbs') -> str:
        args = ', '.join(p.var for p in ty_abs.param)
        if args:
            args = ' ' + args
        body = self.visit_expr(ty_abs.body)
        return body

        return f'ty_lambda{args}: {body}'

    def visit_expr_case(self, case: 'Case') -> str:
        scrut = self.visit_expr(case.scrut)
        line = f'    scrut = {scrut}\n'
        # for alt in case.alts:
        #     alt.Sum_match()
        #     alt.body
        line += '    ...'
        return line

    def visit_expr_let(self, let: 'Block') -> str:
        if self.context.in_expression:
            assignments = ', '.join(f'{b.binder.var} := {self.visit_expr(b.bound)}' for b in let.bindings)
            return f'({assignments}, {self.visit_expr(let.body)})[-1]'
        else:
            with StringIO() as buf:
                for b in let.bindings:
                    buf.write(f'{b.binder.var} = {self.visit_expr(b.bound)}\n')
                buf.write(f'return {self.visit_expr(let.body)}')
                return buf.getvalue()

    def visit_expr_nil_inline(self, nil: 'Expr.Nil') -> str:
        return '()'

    def visit_expr_cons(self, cons: 'Expr.Cons') -> str:
        with StringIO() as buf:
            delim = ''
            buf.write('[')
            for f in cons.front:
                buf.write(delim)
                expr = self.visit_expr(f)
                if ' ' in expr:
                    buf.write('(')
                buf.write(expr)
                if ' ' in expr:
                    buf.write(')')
                delim = ', '

            buf.write(delim)
            expr = self.visit_expr(cons.tail)
            buf.write('*')
            if ' ' in expr:
                buf.write('(')
            buf.write(expr)
            if ' ' in expr:
                buf.write(')')
            buf.write(']')
            return buf.getvalue()

    def visit_expr_update(self, update: 'Update') -> str:
        raise Exception

    def visit_expr_scenario(self, scenario: 'Scenario') -> str:
        raise Exception

    def visit_expr_rec_upd(self, rec_upd: 'Expr.RecUpd') -> str:
        raise Exception

    def visit_expr_tuple_upd(self, tuple_upd: 'Expr.TupleUpd') -> str:
        raise Exception

    def visit_expr_none(self, none: 'Expr.None_') -> str:
        expr = 'None'
        if self.context.in_expression:
            return expr
        else:
            try:
                type_str = self.visit_type(none.type)
                type_str = f'  # type: Optional[{type_str}]'
            except:
                type_str = None

            return f'return {expr}{type_str}'

    def visit_expr_some(self, some: 'Expr.Some') -> str:
        expr = self.visit_expr(some.body)
        if self.in_lambda_body:
            return expr
        else:
            try:
                type_str = self.visit_type(some.type)
                type_str = f'  # type: Optional[{type_str}]'
            except:
                type_str = None

            return f'return {expr}{type_str}'

    def visit_type_var(self, var: 'Type.Var') -> str:
        return var.var

    def visit_type_con(self, con: 'Type.Con') -> str:
        if con.args:
            return f'{con.tycon.full_name}[{", ".join([self.visit_type(a) for a in con.args])}]'
        else:
            return con.tycon.full_name

    def visit_type_prim(self, prim: 'Type.Prim') -> str:
        if PrimType.UNIT == prim.prim:
            return 'Unit'
        elif PrimType.BOOL == prim.prim:
            return 'bool'
        elif PrimType.INT64 == prim.prim:
            return 'int'
        elif PrimType.DECIMAL == prim.prim:
            return 'Decimal'
        elif PrimType.CHAR == prim.prim:
            return 'str'
        elif PrimType.TEXT == prim.prim:
            return 'str'
        elif PrimType.TIMESTAMP == prim.prim:
            return 'datetime'
        elif PrimType.RELTIME == prim.prim:
            return 'timedelta'
        elif PrimType.PARTY == prim.prim:
            return 'Party'
        elif PrimType.LIST == prim.prim:
            if len(prim.args) == 0:
                return 'List'
            else:
                return f'List[{prim.args[0]}]'
        elif PrimType.UPDATE == prim.prim:
            if len(prim.args) == 0:
                return 'Update'
            else:
                return f'Update[{self.visit_type(prim.args[0])}]'
        elif PrimType.SCENARIO == prim.prim:
            if len(prim.args) == 0:
                return 'Scenario'
            else:
                return f'Scenario[{self.visit_type(prim.args[0])}]'
        elif PrimType.DATE == prim.prim:
            return 'Date'
        elif PrimType.CONTRACT_ID == prim.prim:
            return f'ContractId[{self.visit_type(prim.args[0])}]'
        elif PrimType.OPTIONAL == prim.prim:
            if len(prim.args) == 0:
                return 'Optional'
            else:
                return f'Optional[{self.visit_type(prim.args[0])}]'
        elif PrimType.ARROW == prim.prim:
            type_strings = [self.visit_type(a) for a in prim.args]
            return f'Callable[[{", ".join(type_strings[:-1])}], {type_strings[-1]}]'
        elif PrimType.MAP == prim.prim:
            return f'Map[{self.visit_type(prim.args[0])},  {self.visit_type(prim.args[1])}]'
        else:
            raise ValueError(f'unknown Type.Prim: {prim!r}')

    def visit_type_forall(self, forall: 'Type.Forall') -> str:
        # Python does not have a way of modelling existential quantification, so simply throw this
        # information away
        return self.visit_type(forall.body)

    def visit_type_tuple(self, tuple: 'Type.Tuple') -> str:
        return '?'


def python_example_object(*_, **__):
    return '...'



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Mapping, Optional, Sequence, Union

from dazl.model.definition import DamlTemplate
from ._render_base import PrettyPrintBase, decode_special_chars, CodeContext, register_pretty_printer
from ..damlast.daml_lf_1 import Module, Expr, Type, PrimType, DefDataType, PrimCon, PrimLit, \
    ValName, BuiltinFunction, Block, Case, Update, Scenario
from ..damlast.util import pack_arrow_type, unpack_arrow_type


class CSharpPrettyPrint(PrettyPrintBase):

    syntax = 'csharp'

    def lexer(self):
        try:
            from pygments.lexers.dotnet import CSharpLexer
        except ImportError:
            return None
        return CSharpLexer()

    def visit_module(self, module: 'Module') -> str:
        prefix = f'namespace {".".join(module.name.segments[:-1])} {{\n' + \
                 f'public static class {module.name.segments[-1]} {{\n'

        lines = []
        for value in module.values:
            proposed_name = '.'.join(value.name_with_type.name)
            final_name = decode_special_chars(proposed_name)
            if final_name != proposed_name:
                final_name = f'{proposed_name} ({final_name})'

            lines.append('/*')
            lines.append(' NAME: ' + final_name)
            lines.append(' TYPE: ' + repr(value.name_with_type.type))
            lines.append(' EXPR: ' + repr(value.expr))
            lines.append('*/')
            child = self.with_decl(value.name_with_type.name, value.name_with_type.type)
            expr_text = child.visit_expr(value.expr)
            if expr_text is None:
                type_text = self.visit_type(value.name_with_type.type)
                expr_text = f'// {value.expr!r}\npublic extern {type_text} {".".join(value.name_with_type.name)};'
            lines.extend(expr_text.splitlines())
            lines.append('')

        for data_type in module.data_types:
            def_data_text = self.visit_def_data_type(data_type)
            lines.extend(def_data_text.splitlines())
            lines.append('')

        return prefix + '\n'.join('    ' + line.replace('$', '_') if line else '' for line in lines) + '}}\n\n'

    def visit_def_data_type(self, def_data_type: 'DefDataType') -> 'str':
        lines = []
        type_params = [v.var for v in def_data_type.params]
        type_param_text = ('<' + ', '.join(type_params) + '>') if type_params else ''
        if def_data_type.serializable:
            lines.append('[Serializable]')
        if def_data_type.record is not None:
            fields = {f.field: self.visit_type(f.type) for f in def_data_type.record.fields}
            lines.append('public sealed class ' + '.'.join(def_data_type.name.segments) + type_param_text + ' {')
            for name, type_str in fields.items():
                lines.append(f'    public {type_str} {name} {{ get; }} = {name};')
            lines.append('}')
        elif def_data_type.variant is not None:
            fields = {f.field: self.visit_type(f.type) for f in def_data_type.variant.fields}
            lines.append('public sealed class ' + '.'.join(def_data_type.name.segments) + type_param_text + ' {')
            for name, type_str in fields.items():
                lines.append(f'    public {type_str} {name} {{ get; }} = {name};')
            lines.append('}')
        else:
            lines.append('public sealed class ' + '.'.join(def_data_type.name.segments) + '{}')
        return '\n'.join(lines) + '\n'

    def visit_daml_template(self, template: 'DamlTemplate'):

        pass

    def visit_expr_abs_decl_eq(self, expr: 'Expr.Abs'):
        return f"""
        /**
         * Originally defined at ...
         */
        public static boolean operator ==() {{
        }}"""


    def visit_expr_var(self, var: str) -> 'str':
        return self._render_rhs(var, None)

    def visit_expr_val(self, val: 'ValName') -> 'str':
        if self.context.decl_name is not None:
            lines = []
            if self.context.type_abs is not None:
                lines.append(f'// WARNING: Not sure how to render these: {self.context.type_abs}')
            lines.append(f'public static {self.visit_type(self.context.decl_type)} '
                         f'{".".join(self.context.decl_name)} = {val.full_name}')
            return '\n'.join(lines)
        else:
            return self._render_rhs(val.full_name, None)

    def visit_expr_builtin(self, builtin: 'BuiltinFunction') -> 'str':
        return self._render_rhs(f'BuiltIns.{builtin.name}', None)

    def visit_expr_prim_con(self, prim_con: 'PrimCon') -> 'str':
        if PrimCon.CON_TRUE == prim_con:
            text = 'true'
        elif PrimCon.CON_FALSE == prim_con:
            text = 'false'
        elif PrimCon.CON_UNIT == prim_con:
            text = 'Unit'
        else:
            text = repr(prim_con)
        return self._render_rhs(text, None)

    def visit_expr_prim_lit(self, prim_lit: 'PrimLit') -> 'str':
        if prim_lit.int64 is not None:
            text = str(prim_lit.int64)
        elif prim_lit.party is not None:
            text = f'Party("{prim_lit.party}")'
        elif prim_lit.text is not None:
            import json
            text = json.dumps(prim_lit.text)
        elif prim_lit.date is not None:
            from datetime import date, timedelta
            d = date(1970, 1, 1) + timedelta(days=prim_lit.date)
            text = f'Date({d.year}, {d.month}, {d.day})'
        elif prim_lit.timestamp is not None:
            text = f'ConstructDateTime({prim_lit.timestamp})'
        else:
            text = repr(prim_lit)

        return self._render_rhs(text, None)

    def _render_rhs(self, ctor: str, args: Optional[Sequence[str]]):
        if self.context.decl_name is not None:
            lines = []
            if self.context.type_abs is not None:
                lines.append(f'// WARNING: Not sure how to render these: {self.context.type_abs}')

            decl_line = f'public static {self.visit_type(self.context.decl_type)} ' \
                f'{".".join(self.context.decl_name)} = {ctor}'
            if args is not None:
                decl_line += '('
                delim = ''
                for arg in args:
                    decl_line += delim
                    delim = ','
                    decl_line += '\n    '
                    decl_line += arg
                decl_line += ')'
            decl_line += ';'
            return decl_line
        else:
            decl_line = ctor
            if args is not None:
                delim = '('
                for arg in args:
                    decl_line += delim
                    delim = ', '
                    decl_line += arg
                decl_line += ')'

        return decl_line if self.context.in_expression else f'return {decl_line};'

    def visit_expr_rec_con(self, rec_con: 'Expr.RecCon') -> 'str':
        scope = self.with_expression()
        args = [f'{field.field}: {scope.visit_expr(field.expr)}' for field in rec_con.fields]
        return self._render_rhs('new ' + self.visit_type_con(rec_con.tycon), args)

    def visit_expr_rec_proj(self, rec_proj: 'Expr.RecProj') -> 'str':
        scope = self.with_expression()
        return self._render_rhs(f'{scope.visit_expr(rec_proj.record)}.{rec_proj.field}', None)

    def visit_expr_variant_con(self, variant_con: 'Expr.VariantCon') -> 'str':
        scope = self.with_expression()
        args = [f'{variant_con.variant_con}: {scope.visit_expr(variant_con.variant_arg)}']
        return self._render_rhs(self.visit_type_con(variant_con.tycon), args)

    def visit_expr_tuple_con(self, tuple_con: 'Expr.TupleCon') -> 'str':
        scope = self.with_expression()
        args = [f'{field}: {scope.visit_expr(field.expr)}' for field in tuple_con.fields]
        return self._render_rhs('', args)

    def visit_expr_tuple_proj(self, tuple_proj: 'Expr.TupleProj') -> 'str':
        scope = self.with_expression()
        return self._render_rhs(f'{scope.visit_expr(tuple_proj.tuple)}.{tuple_proj.field}', None)

    def visit_expr_app(self, app: 'Expr.App') -> 'str':
        scope = self.with_expression()
        ctor_name = scope.visit_expr(app.fun)
        if self.context.decl_name is not None:
            lines = []
            if self.context.type_abs is not None:
                lines.append(f'// WARNING: Not sure how to render these: {self.context.type_abs}')

            if self.context.type_app:
                type_app_str = '<' + ', '.join(self.visit_type(t) for t in self.context.type_app) + '>'
            else:
                type_app_str = ''
            decl_line = f'public static {self.visit_type(self.context.decl_type)} ' \
                f'{".".join(self.context.decl_name)} = {ctor_name}{type_app_str}('
            delim = ''
            for arg in app.args:
                decl_line += delim
                delim = ','
                decl_line += '\n    '
                decl_line += scope.visit_expr(arg)
            decl_line += ');'

            return decl_line
        else:
            decl_line = ctor_name
            if ' ' in decl_line:
                decl_line = '(' + decl_line + ')'
            delim = '('
            for arg in app.args:
                decl_line += delim
                delim = ', '
                decl_line += scope.visit_expr(arg)
            decl_line += ')'

            return decl_line if self.context.in_expression else f'return {decl_line};'

    def visit_expr_abs_decl(self, abs_: 'Expr.Abs', context: 'CodeContext') -> 'str':
        fn_types = unpack_arrow_type(context.decl_type)

        if context.type_abs is not None:
            generic_args = '<' + (', '.join(self.visit_type_var(tvwk.var) for tvwk in context.type_abs)) + '>'
        else:
            generic_args = ''

        length_of_abstraction = len(abs_.param)
        length_of_proposed_signature = len(fn_types) - 1

        fn_args = {}  # type: Mapping[str, Type]
        ret_type = None  # type: Type
        if length_of_abstraction <= length_of_proposed_signature:
            # TODO: vwt.type and type should always agree with each other; might be nice if we validated that
            #  here instead of just assuming
            fn_args = {vwt.var: self.visit_type(type) for vwt, type in zip(abs_.param, fn_types)}
            ret_type = pack_arrow_type(fn_types[len(abs_.param):])
        elif length_of_abstraction > length_of_proposed_signature:
            # we're passing additional parameters to function returned from the original expression
            fn_args = {vwt.var: self.visit_type(type) for vwt, type in zip(abs_.param, fn_types)}
            return '???'

        fn_args_text = ', '.join(f'{arg_type_text} {arg_name}' for arg_name, arg_type_text in fn_args.items())

        child = self.with_statement_block()
        return_type_text = self.visit_type(ret_type)
        l1 = f'public static {return_type_text} {".".join(context.decl_name)}{generic_args}({fn_args_text}) {{'
        l2 = child.visit_expr(abs_.body) or '...'
        l3 = '}'
        return '\n'.join([l1, *('    ' + text for text in l2.splitlines()), l3])

    def visit_expr_abs_inline(self, abs_: 'Expr.Abs') -> 'str':
        type_params = []
        for param in abs_.param:
            type_expr = self.visit_type(param.type) + ' '
            if type_expr == 'var ':
                type_expr = ''
            type_params.append(type_expr + param.var)
        body_expr = self.visit_expr(abs_.body)
        expr_text = f'({", ".join(type_params)}) => {body_expr}'
        return expr_text if self.context.in_expression else f'return {expr_text};'

    def visit_expr_ty_abs(self, ty_abs: 'Expr.TyAbs') -> 'str':
        return self.with_type_abs(ty_abs.param).visit_expr(ty_abs.body)

    def visit_expr_case(self, case: 'Case') -> 'str':
        pass

    def visit_expr_let(self, let: 'Block') -> 'str':
        lines = []
        scope = self.with_expression()
        for binding in let.bindings:
            type_str = scope.visit_type(binding.binder.type)
            expr_str = scope.visit_expr(binding.bound)
            lines.append(f'{type_str} {binding.binder.var} = {expr_str};')
        lines.append('return ' + scope.visit_expr(let.body) + ';')
        return '\n'.join(lines)

    def visit_expr_nil(self, nil: 'Expr.Nil') -> 'str':
        return self._render_rhs(f'ImmutableList<{self.visit_type(nil.type)}>.Empty', None)

    def visit_expr_cons(self, cons: 'Expr.Cons') -> 'str':
        scope = self.with_expression()
        front_args = ', '.join(scope.visit_expr(f) for f in cons.front)
        tail_arg = scope.visit_expr(cons.tail)
        expr_text = f'ImmutableList.Create<{scope.visit_type(cons.type)}>({front_args}).AddRange({tail_arg})'
        return self._render_rhs(expr_text, None)

    def visit_expr_update(self, update: 'Update') -> 'str':
        pass

    def visit_expr_scenario(self, scenario: 'Scenario') -> 'str':
        pass

    def visit_expr_rec_upd(self, rec_upd: 'Expr.RecUpd') -> 'str':
        pass

    def visit_expr_tuple_upd(self, tuple_upd: 'Expr.TupleUpd') -> 'str':
        pass

    def visit_expr_none(self, none: 'Expr.None_') -> 'str':
        expr_text = f'ImmutableList<{self.visit_type(none.type)}>.Empty'
        return expr_text if self.context.in_expression else f'return {expr_text};'

    def visit_expr_some(self, some: 'Expr.Some') -> 'str':
        return self.visit_expr(some.body)

    def visit_type_var(self, var: 'Union[str, Type.Var]') -> 'str':
        rendered_var = super(CSharpPrettyPrint, self).visit_type_var(var)
        return rendered_var[0].upper() + rendered_var[1:]

    def visit_type_con(self, con: 'Type.Con') -> 'str':
        type_args = [self.visit_type(arg) for arg in con.args]
        if type_args:
            return f'{con.tycon.full_name}<{", ".join(type_args)}>'
        else:
            return con.tycon.full_name

    def visit_type_prim(self, prim: 'Type.Prim') -> 'str':
        if PrimType.UNIT == prim.prim:
            return 'Unit'
        elif PrimType.BOOL == prim.prim:
            return 'boolean'
        elif PrimType.INT64 == prim.prim:
            return 'int'
        elif PrimType.DECIMAL == prim.prim:
            return 'Decimal'
        elif PrimType.CHAR == prim.prim:
            return 'string'
        elif PrimType.TEXT == prim.prim:
            return 'string'
        elif PrimType.TIMESTAMP == prim.prim:
            return 'DateTime'
        elif PrimType.RELTIME == prim.prim:
            return 'timedelta'
        elif PrimType.PARTY == prim.prim:
            return 'Party'
        elif PrimType.LIST == prim.prim:
            if len(prim.args) == 0:
                return 'List<>'
            else:
                return f'List<{self.visit_type(prim.args[0])}>'
        elif PrimType.UPDATE == prim.prim:
            if len(prim.args) == 0:
                return 'Update<>'
            else:
                return f'Update<{self.visit_type(prim.args[0])}>'
        elif PrimType.SCENARIO == prim.prim:
            if len(prim.args) == 0:
                return 'Scenario'
            else:
                return f'Scenario<{self.visit_type(prim.args[0])}>'
        elif PrimType.DATE == prim.prim:
            return 'Date'
        elif PrimType.CONTRACT_ID == prim.prim:
            return f'ContractId<{self.visit_type(prim.args[0])}>'
        elif PrimType.OPTIONAL == prim.prim:
            if len(prim.args) == 0:
                return 'Optional<>'
            else:
                return f'Optional<{self.visit_type(prim.args[0])}>'
        elif PrimType.ARROW == prim.prim:
            type_strings = [self.visit_type(a) for a in prim.args]
            return f'Func<{", ".join(type_strings)}>'
        elif PrimType.MAP == prim.prim:
            return f'Dictionary<{self.visit_type(prim.args[0])},  {self.visit_type(prim.args[1])}>'
        else:
            raise ValueError(f'unknown Type.Prim: {prim!r}')

    def visit_type_forall(self, forall: 'Type.Forall') -> 'str':
        return self.visit_type(forall.body)

    def visit_type_tuple(self, tuple: 'Type.Tuple') -> 'str':
        args = ', '.join(f'{self.visit_type(fwt.type)} {fwt.field}' for fwt in tuple.fields)
        return f'({args})'

    def _visit_type_default(self):
        return 'var'



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import command_submission_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__submission__service__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class CommandSubmissionServiceStub(object):
  """Allows clients to attempt advancing the ledger's state by submitting commands.
  The final states of their submissions are disclosed by the Command Completion Service.
  The on-ledger effects of their submissions are disclosed by the Transaction Service.
  Commands may fail in 4 distinct manners:

  1) ``INVALID_PARAMETER`` gRPC error on malformed payloads and missing required fields.
  2) Failure communicated in the gRPC error.
  3) Failure communicated in a Completion.
  4) A Checkpoint with ``record_time`` > command ``mrt`` arrives through the Completion Stream, and the command's Completion was not visible before. In this case the command is lost.

  Clients that do not receive a successful completion about their submission MUST NOT assume that it was successful.
  Clients SHOULD subscribe to the CompletionStream before starting to submit commands to prevent race conditions.

  Interprocess tracing of command submissions may be achieved via Zipkin by filling out the ``trace_context`` field.
  The server will return a child context of the submitted one, (or a new one if the context was missing) on both the Completion and Transaction streams.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Submit = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.CommandSubmissionService/Submit',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__submission__service__pb2.SubmitRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )


class CommandSubmissionServiceServicer(object):
  """Allows clients to attempt advancing the ledger's state by submitting commands.
  The final states of their submissions are disclosed by the Command Completion Service.
  The on-ledger effects of their submissions are disclosed by the Transaction Service.
  Commands may fail in 4 distinct manners:

  1) ``INVALID_PARAMETER`` gRPC error on malformed payloads and missing required fields.
  2) Failure communicated in the gRPC error.
  3) Failure communicated in a Completion.
  4) A Checkpoint with ``record_time`` > command ``mrt`` arrives through the Completion Stream, and the command's Completion was not visible before. In this case the command is lost.

  Clients that do not receive a successful completion about their submission MUST NOT assume that it was successful.
  Clients SHOULD subscribe to the CompletionStream before starting to submit commands to prevent race conditions.

  Interprocess tracing of command submissions may be achieved via Zipkin by filling out the ``trace_context`` field.
  The server will return a child context of the submitted one, (or a new one if the context was missing) on both the Completion and Transaction streams.
  """

  def Submit(self, request, context):
    """Submit a single composite command.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_CommandSubmissionServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Submit': grpc.unary_unary_rpc_method_handler(
          servicer.Submit,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__submission__service__pb2.SubmitRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.CommandSubmissionService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Object model for data capture.
"""

from typing import Dict, Iterator, Optional

from ...model.core import ContractId, ContractData
from ...model.types_store import PackageStore


class LedgerCapture:
    """
    State built up during a ledger run.
    """
    def __init__(self):
        self.entries = dict()  # type: Dict[ContractId, LedgerCaptureEntry]
        self.store: Optional[PackageStore] = None

    def capture(self, party: str, contract_id: ContractId, contract_data: Optional[ContractData]):
        entry = self.entries.get(contract_id)
        if entry is not None:
            entry.extend(party, contract_id, contract_data)
        else:
            self.entries[contract_id] = LedgerCaptureEntry(party, contract_id, contract_data)

    def capture_archive(self, party: str, contract_id: ContractId):
        return self.capture(party, contract_id, None)

    def excluding_inactive(self):
        return filter(lambda e: not e.is_archived(), self)

    def filtered_by(self, predicate):
        return filter(predicate, self)

    def __iter__(self) -> 'Iterator[LedgerCaptureEntry]':
        return iter(self.entries.values())

    def __repr__(self):
        return '<LedgerCapture(entry_count={!r})>'.format(len(self.entries))


class LedgerCaptureEntry:
    def __init__(self, party, contract_id, contract_data):
        self.parties = {party: True}
        self.contract_id = contract_id
        self.template_id = contract_id.template_id
        self.contract_args = contract_data
        self.errors = []

    def extend(self, party, _, contract_data):
        active = contract_data is not None
        self.parties[party] = active

        if active:
            # make sure this party is seeing a consistent view of the same contract
            pass

    def is_archived(self):
        return all(not active for active in self.parties.values())

    def contract_state(self, party):
        p = self.parties.get(party)
        if p is not None:
            return "CREATED" if p else "ARCHIVED"
        else:
            return None

    def __repr__(self):
        return '<LedgerEntry(parties={}, contract_id={}, template_id={}, contract_args={})>'.format(
            self.parties, self.contract_id, self.template_id, self.contract_args)




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import transaction_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2


class TransactionServiceStub(object):
  """Allows clients to read transactions from the ledger.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetTransactions = channel.unary_stream(
        '/com.digitalasset.ledger.api.v1.TransactionService/GetTransactions',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsResponse.FromString,
        )
    self.GetTransactionTrees = channel.unary_stream(
        '/com.digitalasset.ledger.api.v1.TransactionService/GetTransactionTrees',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionTreesResponse.FromString,
        )
    self.GetTransactionByEventId = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.TransactionService/GetTransactionByEventId',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.FromString,
        )
    self.GetTransactionById = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.TransactionService/GetTransactionById',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.FromString,
        )
    self.GetFlatTransactionByEventId = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.TransactionService/GetFlatTransactionByEventId',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.FromString,
        )
    self.GetFlatTransactionById = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.TransactionService/GetFlatTransactionById',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.FromString,
        )
    self.GetLedgerEnd = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.TransactionService/GetLedgerEnd',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndResponse.FromString,
        )


class TransactionServiceServicer(object):
  """Allows clients to read transactions from the ledger.
  """

  def GetTransactions(self, request, context):
    """Read the ledger's filtered transaction stream for a set of parties.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransactionTrees(self, request, context):
    """Read the ledger's complete transaction tree stream for a set of parties.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransactionByEventId(self, request, context):
    """Lookup a transaction tree by the ID of an event that appears within it.
    Returns ``NOT_FOUND`` if no such transaction exists.
    For looking up a transaction instead of a transaction tree, please see GetFlatTransactionByEventId
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransactionById(self, request, context):
    """Lookup a transaction tree by its ID.
    Returns ``NOT_FOUND`` if no such transaction exists.
    For looking up a transaction instead of a transaction tree, please see GetFlatTransactionById
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetFlatTransactionByEventId(self, request, context):
    """Lookup a transaction by the ID of an event that appears within it.
    Returns ``NOT_FOUND`` if no such transaction exists.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetFlatTransactionById(self, request, context):
    """Lookup a transaction by its ID.
    Returns ``NOT_FOUND`` if no such transaction exists.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetLedgerEnd(self, request, context):
    """Get the current ledger end.
    Subscriptions started with the returned offset will serve transactions created after this RPC was called.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_TransactionServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetTransactions': grpc.unary_stream_rpc_method_handler(
          servicer.GetTransactions,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsResponse.SerializeToString,
      ),
      'GetTransactionTrees': grpc.unary_stream_rpc_method_handler(
          servicer.GetTransactionTrees,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionsRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionTreesResponse.SerializeToString,
      ),
      'GetTransactionByEventId': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransactionByEventId,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.SerializeToString,
      ),
      'GetTransactionById': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransactionById,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionResponse.SerializeToString,
      ),
      'GetFlatTransactionByEventId': grpc.unary_unary_rpc_method_handler(
          servicer.GetFlatTransactionByEventId,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByEventIdRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.SerializeToString,
      ),
      'GetFlatTransactionById': grpc.unary_unary_rpc_method_handler(
          servicer.GetFlatTransactionById,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetTransactionByIdRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetFlatTransactionResponse.SerializeToString,
      ),
      'GetLedgerEnd': grpc.unary_unary_rpc_method_handler(
          servicer.GetLedgerEnd,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__service__pb2.GetLedgerEndResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.TransactionService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/testing/reset_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/testing/reset_service.proto',
  package='com.digitalasset.ledger.api.v1.testing',
  syntax='proto3',
  serialized_options=_b('\n&com.digitalasset.ledger.api.v1.testingB\026ResetServiceOuterClass'),
  serialized_pb=_b('\n:com/digitalasset/ledger/api/v1/testing/reset_service.proto\x12&com.digitalasset.ledger.api.v1.testing\x1a\x1bgoogle/protobuf/empty.proto\"!\n\x0cResetRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t2e\n\x0cResetService\x12U\n\x05Reset\x12\x34.com.digitalasset.ledger.api.v1.testing.ResetRequest\x1a\x16.google.protobuf.EmptyB@\n&com.digitalasset.ledger.api.v1.testingB\x16ResetServiceOuterClassb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_RESETREQUEST = _descriptor.Descriptor(
  name='ResetRequest',
  full_name='com.digitalasset.ledger.api.v1.testing.ResetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.testing.ResetRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=131,
  serialized_end=164,
)

DESCRIPTOR.message_types_by_name['ResetRequest'] = _RESETREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ResetRequest = _reflection.GeneratedProtocolMessageType('ResetRequest', (_message.Message,), dict(
  DESCRIPTOR = _RESETREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.testing.reset_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.testing.ResetRequest)
  ))
_sym_db.RegisterMessage(ResetRequest)


DESCRIPTOR._options = None

_RESETSERVICE = _descriptor.ServiceDescriptor(
  name='ResetService',
  full_name='com.digitalasset.ledger.api.v1.testing.ResetService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=166,
  serialized_end=267,
  methods=[
  _descriptor.MethodDescriptor(
    name='Reset',
    full_name='com.digitalasset.ledger.api.v1.testing.ResetService.Reset',
    index=0,
    containing_service=None,
    input_type=_RESETREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_RESETSERVICE)

DESCRIPTOR.services_by_name['ResetService'] = _RESETSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/completion.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import trace_context_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2
from google.rpc import status_pb2 as google_dot_rpc_dot_status__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/completion.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B\024CompletionOuterClass'),
  serialized_pb=_b('\n/com/digitalasset/ledger/api/v1/completion.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a\x32\x63om/digitalasset/ledger/api/v1/trace_context.proto\x1a\x17google/rpc/status.proto\"\xa2\x01\n\nCompletion\x12\x12\n\ncommand_id\x18\x01 \x01(\t\x12\"\n\x06status\x18\x02 \x01(\x0b\x32\x12.google.rpc.Status\x12\x16\n\x0etransaction_id\x18\x03 \x01(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContextB6\n\x1e\x63om.digitalasset.ledger.api.v1B\x14\x43ompletionOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,google_dot_rpc_dot_status__pb2.DESCRIPTOR,])




_COMPLETION = _descriptor.Descriptor(
  name='Completion',
  full_name='com.digitalasset.ledger.api.v1.Completion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='command_id', full_name='com.digitalasset.ledger.api.v1.Completion.command_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='status', full_name='com.digitalasset.ledger.api.v1.Completion.status', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='com.digitalasset.ledger.api.v1.Completion.transaction_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.Completion.trace_context', index=3,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=161,
  serialized_end=323,
)

_COMPLETION.fields_by_name['status'].message_type = google_dot_rpc_dot_status__pb2._STATUS
_COMPLETION.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
DESCRIPTOR.message_types_by_name['Completion'] = _COMPLETION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Completion = _reflection.GeneratedProtocolMessageType('Completion', (_message.Message,), dict(
  DESCRIPTOR = _COMPLETION,
  __module__ = 'com.digitalasset.ledger.api.v1.completion_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.Completion)
  ))
_sym_db.RegisterMessage(Completion)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains miscellaneous utility methods that don't really fit anywhere else.
"""

from typing import Generator, Tuple, TypeVar, Iterable

T = TypeVar('T')


def boundary_iter(obj: Iterable[T]) -> Generator[Tuple[bool, T], None, None]:
    """
    Iterates over an iterable, returning a boolean and the underlying values indicating
    whether the iterator has reached its last item.

    >>> list(boundary_iter(range(0, 3)))
    [(False, 0), (False, 1), (True, 2)]

    >>> list(boundary_iter('A'))
    [(True, 'A')]

    # empty example
    >>> list(boundary_iter([]))
    []

    # iterating over an infinite iterable
    >>> import itertools
    >>> list(itertools.islice(boundary_iter(iter(lambda:0,1)), 4))
    [(False, 0), (False, 0), (False, 0), (False, 0)]
    """
    gen = iter(obj)
    prev_entry = None
    while True:
        try:
            entry = next(gen)
            if prev_entry is not None:
                yield (False, prev_entry)
            prev_entry = entry
        except StopIteration:
            if prev_entry is not None:
                yield (True, prev_entry)
            break


def flatten(obj):
    ret = []
    for sublist in obj:
        ret.extend(sublist)
    return ret



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import package_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2


class PackageServiceStub(object):
  """Allows clients to query the DAML-LF packages that are supported by the server.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ListPackages = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.PackageService/ListPackages',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesResponse.FromString,
        )
    self.GetPackage = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.PackageService/GetPackage',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageResponse.FromString,
        )
    self.GetPackageStatus = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.PackageService/GetPackageStatus',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusResponse.FromString,
        )


class PackageServiceServicer(object):
  """Allows clients to query the DAML-LF packages that are supported by the server.
  """

  def ListPackages(self, request, context):
    """Returns the identifiers of all supported packages.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPackage(self, request, context):
    """Returns the contents of a single package, or a ``NOT_FOUND`` error if the requested package is unknown.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPackageStatus(self, request, context):
    """Returns the status of a single package.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PackageServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ListPackages': grpc.unary_unary_rpc_method_handler(
          servicer.ListPackages,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesResponse.SerializeToString,
      ),
      'GetPackage': grpc.unary_unary_rpc_method_handler(
          servicer.GetPackage,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageResponse.SerializeToString,
      ),
      'GetPackageStatus': grpc.unary_unary_rpc_method_handler(
          servicer.GetPackageStatus,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.PackageService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import asyncio
import warnings
from asyncio import Future
from collections import defaultdict
from typing import Awaitable, Dict, List, Optional, Union, Collection, cast

from ..client._reader_match import is_match
from ..model.core import ContractId, ContractsState, ContractMatch, \
    UnknownTemplateWarning, ContractContextualData, ContractContextualDataCollection
from ..model.reading import ContractCreateEvent, ContractArchiveEvent
from ..model.types import TypeReference
from ..model.types_store import PackageStore
from ..util.asyncio_util import ContextFreeFuture, safe_create_future, propagate, completed, await_then


class ActiveContractSet:
    def __init__(self):
        self.metadata_future = ContextFreeFuture()
        self._tcdata = defaultdict(TemplateContractData)  # type: Dict[TypeReference, TemplateContractData]

    def handle_create(self, event: ContractCreateEvent) -> None:
        self._tcdata[event.cid.template_id].handle_create(event)

    def handle_archive(self, event: ContractArchiveEvent) -> None:
        self._tcdata[event.cid.template_id].handle_archive(event)

    def get(self, cid: 'Union[str, ContractId]') -> 'Optional[ContractContextualData]':
        """
        Return information for the associated :class:`ContractId`.

        :param cid: Either a ContractId or a string that represents a ContractId.
        :return: Contract information for the specified contract ID.
        """
        tcd = self._tcdata.get(cid.template_id) if isinstance(cid, ContractId) else None
        if tcd is not None:
            return tcd.get(cid)
        else:
            # untyped ContractId lookup, so scan across all templates
            for tcd in self._tcdata.values():
                cxd = tcd.get(cid)
                if cxd is not None:
                    return cxd
            return None

    def read_active(self, template_name: str, match: ContractMatch = None) -> ContractsState:
        return {cxd.cid: cxd.cdata for cxd in self.read_full(template_name, match)}

    def read_full(self,
                  template_name: str,
                  match: ContractMatch = None,
                  include_archived: bool = False) -> ContractContextualDataCollection:
        if not self.metadata_future.done():
            return ContractContextualDataCollection()

        unfiltered = self._get_template_state(self.metadata_future.result(), template_name)
        all_items = [item
                     for tcd in unfiltered.values()
                     for item in tcd.subset(match, include_archived)]
        return ContractContextualDataCollection(all_items)

    def read_async(self, template_name: str, match: ContractMatch = None, min_count: int = 1) \
            -> Awaitable[ContractsState]:
        if self.metadata_future.done():
            unfiltered = self._get_template_state(self.metadata_future.result(), template_name)
            if len(unfiltered) > 1:
                warnings.warn('Wildcard searches are not supported on async ACS queries',
                              UnknownTemplateWarning, stacklevel=3)
                unfiltered = dict()  # type: Dict[TypeReference, TemplateContractData]

            if len(unfiltered) == 0:
                # TODO: A slightly smarter implementation could hang around until/if a matching
                #  template shows up
                return completed({})

            (tt, tcd), = unfiltered.items()

            query = PendingQuery(asyncio.get_event_loop(), match, min_count)
            # if the current state is already a match, then don't remember the query since we're
            # already done
            if tcd is None or not query.check_ready(tcd):
                tcd.register_query(query)

            return await_then(query.future, lambda cxds: {cxd.cid: cxd.cdata for cxd in cxds})
        else:
            # delay the invocation of this entire method call until metadata is made available to us
            future = safe_create_future()

            def delayed_invoke(_):
                propagate(
                    asyncio.ensure_future(self.read_async(template_name, match, min_count)), future)

            self.metadata_future.add_done_callback(delayed_invoke)
            return future

    def _get_template_state(self, store: PackageStore, template_name: str) -> \
            'Dict[TypeReference, TemplateContractData]':
        matching_templates = store.resolve_template_type(template_name)
        if matching_templates:
            return {tt: self._tcdata[tt] for tt in matching_templates}
        else:
            warnings.warn(f'Unknown template name: {template_name}', UnknownTemplateWarning,
                          stacklevel=4)
            return {}


class TemplateContractData:
    """
    Storage for state related to a specific :class:`Template`.
    """
    def __init__(self):
        # raw ContractId to ContractContextualData
        self._data = dict()  # type: Dict[str, ContractContextualData]
        # outstanding ACS queries
        self._queries = list()  # type: List[PendingQuery]

    def handle_create(self, event: ContractCreateEvent) -> None:
        self._data[event.cid.contract_id] = ContractContextualData(
            cid=event.cid,
            cdata=event.cdata,
            effective_at=event.time,
            archived_at=None,
            active=True)

        self._queries = [q for q in self._queries if not q.check_ready(self)]

    def handle_archive(self, event: ContractArchiveEvent) -> None:
        self._data[event.cid.contract_id] = ContractContextualData(
            cid=event.cid,
            cdata=None,
            effective_at=self._data[event.cid.contract_id].effective_at,
            archived_at=event.time,
            active=False)

    def get(self, cid: 'Union[str, ContractId]') -> 'Optional[ContractContextualData]':
        if isinstance(cid, ContractId):
            cid = cid.contract_id
        return self._data.get(cid)

    def subset(self, match: 'ContractMatch', include_archived: bool) -> 'Collection[ContractContextualData]':
        return [cxd for cxd in self._data.values()
                if (include_archived or cxd.active)
                and is_match(match, cxd.cdata)]

    def register_query(self, query: 'PendingQuery') -> None:
        self._queries.append(query)


class PendingQuery:
    def __init__(self, loop, match, min_count: int):
        self.future = loop.create_future()  # type: Awaitable[Collection[ContractContextualData]]
        self.match = match
        self.min_count = min_count

    def check_ready(self, tcd: 'TemplateContractData') -> bool:
        matches = tcd.subset(self.match, False)
        if len(matches) >= self.min_count:
            cast(Future, self.future).set_result(matches)
            return True
        else:
            return False



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ...._gen.com.digitalasset.ledger.api.v1.commands_pb2 import Command, CreateCommand, ExerciseCommand
from ...._gen.com.digitalasset.ledger.api.v1.command_service_pb2_grpc import CommandServiceStub
from ...._gen.com.digitalasset.ledger.api.v1.command_submission_service_pb2 import SubmitRequest
from ...._gen.com.digitalasset.ledger.api.v1.commands_pb2 import Commands
from ...._gen.com.digitalasset.ledger.api.v1.ledger_identity_service_pb2 import \
    GetLedgerIdentityRequest
from ...._gen.com.digitalasset.ledger.api.v1.ledger_identity_service_pb2_grpc import \
    LedgerIdentityServiceStub
from ...._gen.com.digitalasset.ledger.api.v1.ledger_offset_pb2 import LedgerOffset
from ...._gen.com.digitalasset.ledger.api.v1.package_service_pb2 import ListPackagesRequest, \
    GetPackageRequest
from ...._gen.com.digitalasset.ledger.api.v1.package_service_pb2_grpc import PackageServiceStub
from ...._gen.com.digitalasset.ledger.api.v1.testing.time_service_pb2 import GetTimeRequest, SetTimeRequest
from ...._gen.com.digitalasset.ledger.api.v1.testing.time_service_pb2_grpc import TimeServiceStub
from ...._gen.com.digitalasset.ledger.api.v1.transaction_filter_pb2 import Filters, \
    TransactionFilter
from ...._gen.com.digitalasset.ledger.api.v1.transaction_service_pb2 import GetTransactionsRequest, \
    GetLedgerEndRequest
from ...._gen.com.digitalasset.ledger.api.v1.transaction_service_pb2_grpc import \
    TransactionServiceStub
from ...._gen.da.daml_lf_pb2 import ArchivePayload
from ...._gen.da.daml_lf_1_pb2 import DefDataType, PrimType, DefTemplate, TemplateChoice
from ...._gen.com.digitalasset.ledger.api.v1.event_pb2 import Event
from ...._gen.com.digitalasset.ledger.api.v1.transaction_pb2 import Transaction
from ...._gen.com.digitalasset.ledger.api.v1.transaction_service_pb2 import GetTransactionsResponse
from ...._gen.com.digitalasset.ledger.api.v1.value_pb2 import Record, Variant, Value, List



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains internal utilities for managing the lifecycle of a process.
"""

import logging
import subprocess
import sys
import time

from datetime import datetime, timedelta
from pathlib import Path
from threading import Thread
from typing import Any, ContextManager, Optional, Sequence, Union

from .io import is_port_alive


DEFAULT_WATCH_TIMEOUT = timedelta(seconds=10)


class ProcessContext:
    """
    Information about how to launch a process, and a context manager that optionally manages
    any temporary resources that need to be kept around while the process is running.
    """

    def __init__(self, args: 'Sequence[Any]', *,
                 secondary_context: 'Optional[ContextManager]' = None,
                 watch_port: 'Optional[int]' = None,
                 watch_timeout: 'Union[None, int, float, timedelta]' = DEFAULT_WATCH_TIMEOUT,
                 logger: 'Optional[logging.Logger]' = None,
                 working_directory: 'Union[None, str, Path]' = None,
                 extra_data: 'Optional[dict]' = None):
        self.args = [str(arg) for arg in args]
        self.secondary_context = secondary_context
        self.watch_port = watch_port

        if isinstance(watch_timeout, (int, float)):
            self.watch_timeout = timedelta(seconds=watch_timeout)
        elif isinstance(watch_timeout, timedelta):
            self.watch_timeout = watch_timeout
        elif watch_timeout is None:
            self.watch_timeout = None
        else:
            raise TypeError(f'Not a valid watch timeout: {watch_timeout!r}')

        self.watch_timeout = watch_timeout
        self.logger = logger

        if isinstance(working_directory, Path):
            self.working_directory = str(working_directory)
        elif isinstance(working_directory, str):
            self.working_directory = working_directory
        elif working_directory is None:
            self.working_directory = None
        else:
            raise TypeError(f'Not a valid working directory: {working_directory!r}')

        self.extra_data = extra_data or {}

    def __enter__(self):
        if self.secondary_context is not None:
            self.secondary_context.__enter__()
        return self.args

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.secondary_context is not None:
            self.secondary_context.__exit__(exc_type, exc_val, exc_tb)


class ProcessWatcher:
    """
    Manages the lifetime of an external process.
    """
    def __init__(self, process_context: 'ProcessContext'):
        self.process = None
        self.process_context = process_context
        self.logger = self.process_context.logger or logging.getLogger()

        self._start_entered = False

    def start(self) -> None:
        if self._start_entered:
            return
        self._start_entered = True

        self.process_context.__enter__()
        self.process = subprocess.Popen(
            self.process_context.args,
            cwd=self.process_context.working_directory,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True)

        stdout = sys.stdout
        stderr = sys.stderr
        stdout_log_thread = Thread(target=self._stdout_monitor, args=[stdout])
        stdout_log_thread.start()
        stderr_log_thread = Thread(target=self._stderr_monitor, args=[stderr])
        stderr_log_thread.start()

        port = self.process_context.watch_port
        if port is None:
            return

        alive = False
        if self.process_context.watch_timeout is not None:
            max_time_wait = datetime.utcnow() + self.process_context.watch_timeout
        else:
            max_time_wait = None

        logging.debug('Waiting for port %s to be alive on pid %s...', port, self.process.pid)
        while ((max_time_wait is None or datetime.utcnow() < max_time_wait) and
               self.process.poll() is None and not alive):
            alive = is_port_alive(port)
            if not alive:
                time.sleep(0.1)
        if not alive:
            return_code = self.process.returncode
            if return_code is not None:
                from ..model.core import ProcessDiedException
                raise ProcessDiedException(return_code, 'The sandbox exited with an error code')
            raise Exception('Timed out while waiting for the sandbox to start')

    def stop(self) -> None:
        if self.process is not None:
            try:
                import psutil
                children = psutil.Process(pid=self.process.pid).children()
                logging.info('Child processes: %s', children)
            except:
                children = []

            self.process.terminate()
            try:
                # give the process two seconds to die
                self.process.communicate(timeout=5)

            except subprocess.TimeoutExpired:
                logging.error('Had trouble killing a sandbox process normally; '
                              'it will be forcibly killed.')
                # if there are child processes, start working around the parent process and try to
                # terminate them directly
                for child in children:
                    try:
                        child.kill()
                    except:
                        pass

                self.process.kill()
                self.process.communicate()
            self.process = None
        if self.process_context is not None:
            self.process_context.__exit__(None, None, None)
            self.process_context = None

    def run(self) -> int:
        """
        Run the process to completion.
        """
        self.start()
        self.process.wait()
        return self.process.returncode

    def _stdout_monitor(self, stdout):
        sys.stdout = stdout
        # noinspection PyBroadException
        try:
            for line in self.process.stdout:
                self.logger.info(line.rstrip('\n'))
        except:  # noqa
            pass

    def _stderr_monitor(self, stderr):
        sys.stderr = stderr
        # noinspection PyBroadException
        try:
            for line in self.process.stderr:
                # TODO: Make this configurable; happens to be that our tools currently print things
                # to stderr that probably better belong on stdout
                self.logger.info(line.rstrip('\n'))
        except:  # noqa
            pass

    def __enter__(self) -> 'ProcessWatcher':
        self.start()
        return self

    def __exit__(self, typ, value, traceback) -> None:
        self.stop()



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
:mod:`dazl.util` package
========================

Module that exposes general utility functions.
"""

from . import dar, io



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/ledger_identity_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import trace_context_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/ledger_identity_service.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B\037LedgerIdentityServiceOuterClass'),
  serialized_pb=_b('\n<com/digitalasset/ledger/api/v1/ledger_identity_service.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a\x32\x63om/digitalasset/ledger/api/v1/trace_context.proto\"`\n\x18GetLedgerIdentityRequest\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\".\n\x19GetLedgerIdentityResponse\x12\x11\n\tledger_id\x18\x01 \x01(\t2\xa2\x01\n\x15LedgerIdentityService\x12\x88\x01\n\x11GetLedgerIdentity\x12\x38.com.digitalasset.ledger.api.v1.GetLedgerIdentityRequest\x1a\x39.com.digitalasset.ledger.api.v1.GetLedgerIdentityResponseBA\n\x1e\x63om.digitalasset.ledger.api.v1B\x1fLedgerIdentityServiceOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,])




_GETLEDGERIDENTITYREQUEST = _descriptor.Descriptor(
  name='GetLedgerIdentityRequest',
  full_name='com.digitalasset.ledger.api.v1.GetLedgerIdentityRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetLedgerIdentityRequest.trace_context', index=0,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=148,
  serialized_end=244,
)


_GETLEDGERIDENTITYRESPONSE = _descriptor.Descriptor(
  name='GetLedgerIdentityResponse',
  full_name='com.digitalasset.ledger.api.v1.GetLedgerIdentityResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetLedgerIdentityResponse.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=246,
  serialized_end=292,
)

_GETLEDGERIDENTITYREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
DESCRIPTOR.message_types_by_name['GetLedgerIdentityRequest'] = _GETLEDGERIDENTITYREQUEST
DESCRIPTOR.message_types_by_name['GetLedgerIdentityResponse'] = _GETLEDGERIDENTITYRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetLedgerIdentityRequest = _reflection.GeneratedProtocolMessageType('GetLedgerIdentityRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETLEDGERIDENTITYREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.ledger_identity_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetLedgerIdentityRequest)
  ))
_sym_db.RegisterMessage(GetLedgerIdentityRequest)

GetLedgerIdentityResponse = _reflection.GeneratedProtocolMessageType('GetLedgerIdentityResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETLEDGERIDENTITYRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.ledger_identity_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetLedgerIdentityResponse)
  ))
_sym_db.RegisterMessage(GetLedgerIdentityResponse)


DESCRIPTOR._options = None

_LEDGERIDENTITYSERVICE = _descriptor.ServiceDescriptor(
  name='LedgerIdentityService',
  full_name='com.digitalasset.ledger.api.v1.LedgerIdentityService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=295,
  serialized_end=457,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetLedgerIdentity',
    full_name='com.digitalasset.ledger.api.v1.LedgerIdentityService.GetLedgerIdentity',
    index=0,
    containing_service=None,
    input_type=_GETLEDGERIDENTITYREQUEST,
    output_type=_GETLEDGERIDENTITYRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_LEDGERIDENTITYSERVICE)

DESCRIPTOR.services_by_name['LedgerIdentityService'] = _LEDGERIDENTITYSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import package_management_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2


class PackageManagementServiceStub(object):
  """Status: experimental interface, will change before it is deemed production
  ready

  Query the DAML-LF packages supported by the ledger participant and upload
  DAR files. We use 'backing participant' to refer to this specific participant
  in the methods of this API.
  When the participant is run in mode requiring authentication, all the calls 
  in this interface will respond with UNAUTHENTICATED, if the caller fails
  to provide a valid access token, and will respond with PERMISSION_DENIED, if
  the claims in the token are insufficient to perform a given operation.
  Subsequently, only specific errors of individual calls not related to 
  authorization will be described.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ListKnownPackages = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.admin.PackageManagementService/ListKnownPackages',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.ListKnownPackagesRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.ListKnownPackagesResponse.FromString,
        )
    self.UploadDarFile = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.admin.PackageManagementService/UploadDarFile',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.UploadDarFileRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.UploadDarFileResponse.FromString,
        )


class PackageManagementServiceServicer(object):
  """Status: experimental interface, will change before it is deemed production
  ready

  Query the DAML-LF packages supported by the ledger participant and upload
  DAR files. We use 'backing participant' to refer to this specific participant
  in the methods of this API.
  When the participant is run in mode requiring authentication, all the calls 
  in this interface will respond with UNAUTHENTICATED, if the caller fails
  to provide a valid access token, and will respond with PERMISSION_DENIED, if
  the claims in the token are insufficient to perform a given operation.
  Subsequently, only specific errors of individual calls not related to 
  authorization will be described.
  """

  def ListKnownPackages(self, request, context):
    """Returns the details of all DAML-LF packages known to the backing
    participant.
    This request will always succeed.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UploadDarFile(self, request, context):
    """Upload a DAR file to the backing participant.
    Depending on the ledger implementation this might also make the package 
    available on the whole ledger. This call might not be supported by some 
    ledger implementations. Canton could be an example, where uploading a DAR
    is not sufficient to render it usable, it must be activated first.
    This method will return UNIMPLEMENTED, if DAR package uploading is not
    supported by the backing participant. If DAR file is too big or is 
    malformed, the backing participant will respond with INVALID_ARGUMENT.
    The maximum supported size is implementation specific.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PackageManagementServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ListKnownPackages': grpc.unary_unary_rpc_method_handler(
          servicer.ListKnownPackages,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.ListKnownPackagesRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.ListKnownPackagesResponse.SerializeToString,
      ),
      'UploadDarFile': grpc.unary_unary_rpc_method_handler(
          servicer.UploadDarFile,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.UploadDarFileRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_admin_dot_package__management__service__pb2.UploadDarFileResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.admin.PackageManagementService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any

from ..model.core import ContractId
from ..model.types import Type, UnsupportedType, VariantType, RecordType, ListType, \
    ContractIdType, TemplateChoice, TypeEvaluationContext, OptionalType, MapType
from ..model.writing import Command, CreateCommand, ExerciseCommand, AbstractSerializer
from ..util.prim_types import to_int, to_str, to_decimal, to_date, to_datetime, \
    unflatten_dotted_keys, to_hashable


class ValidateError(ValueError):
    pass


class ValidateSerializer(AbstractSerializer[Command, Any]):
    """
    A serializer that doesn't actually write to an on-wire form, but merely does lightweight
    conversion and give on-wire serialization the best chance of succeeding.

    This class essentially handles type conversion before a command is fully created and sent to
    a background thread for eventual writing to the Ledger API.
    """

    def serialize_create_command(self, template_type: RecordType, template_args: Any) \
            -> CreateCommand:
        return CreateCommand(template=template_type, arguments=template_args)

    def serialize_exercise_command(
            self, contract_id: ContractId, choice_info: TemplateChoice, choice_args: Any) \
            -> ExerciseCommand:
        return ExerciseCommand(contract=contract_id, choice=choice_info.name, arguments=choice_args)

    def serialize_unit(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return obj

    def serialize_bool(self, context: TypeEvaluationContext, obj: Any) -> Any:
        from ..protocols.v1.pb_parse_event import to_bool
        return to_bool(obj)

    def serialize_text(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return to_str(obj)

    def serialize_int(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return to_int(obj)

    def serialize_decimal(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return to_decimal(obj)

    def serialize_party(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return to_str(obj)

    def serialize_date(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return to_date(obj)

    def serialize_datetime(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return to_datetime(obj)

    def serialize_timedelta(self, context: TypeEvaluationContext, obj: Any) -> Any:
        return obj

    def serialize_contract_id(self, context: TypeEvaluationContext, tt: ContractIdType, obj: Any) \
            -> Any:
        return obj

    def serialize_optional(self, context: TypeEvaluationContext, tt: OptionalType, obj: Any) -> Any:
        # a `None` object is, of course, a valid Optional
        if obj is None:
            return None

        # any other object is considered in the context of its inner type
        return self._serialize_dispatch(context.append_path('?'), tt.type_parameter, obj)

    def serialize_list(self, context: TypeEvaluationContext, tt: ListType, obj: Any) -> Any:
        # a `None` object is trivially interpreted as an empty list
        if obj is None:
            return []

        if isinstance(obj, str):
            # Strings are a bit special because they're enumerable, but we don't really want to
            # treat them that way here. Additionally, if this is a bare string with commas in it,
            # use those commas to split the string. Additionally, if it's the empty string, treat
            # that as a null list instead of a single list with an empty string inside of it.
            obj = obj.split(',') if obj else []

        from collections import Collection
        if not isinstance(obj, Collection):
            obj = [obj]

        return [self._serialize_dispatch(context.append_path(f'[{i}]'), tt.type_parameter, item)
                for i, item in enumerate(obj)]

    def serialize_map(self, context: TypeEvaluationContext, tt: MapType, obj: Any) -> Any:
        if obj is None:
            return {}

        from collections import Mapping
        if isinstance(obj, Mapping):
            reformatted = {}
            for i, (key, value) in enumerate(obj.items()):
                new_key = to_hashable(self._serialize_dispatch(
                    context.append_path(f'[key {i}]'), tt.key_type, key))
                new_value = to_hashable(self._serialize_dispatch(
                    context.append_path(f'[value {i}]'), tt.value_type, value))
                reformatted[new_key] = new_value
            return reformatted
        else:
            raise ValueError(f'expected a dict here (got {type(obj)}: {obj!r} instead)')

    def serialize_record(self, context: TypeEvaluationContext, tt: RecordType, obj: Any) -> Any:
        from collections.abc import Mapping
        if isinstance(obj, Mapping):
            # pull out any specialized dotted-field mappings
            reformatted = unflatten_dotted_keys(obj)
            missing_keys = set()
            for name, field_type in tt.named_args:
                if name in reformatted:
                    new_value = self._serialize_dispatch(
                        context.append_path(name), field_type, reformatted[name])
                    reformatted[name] = new_value
                else:
                    missing_keys.add(name)
            if missing_keys:
                raise ValueError(f'missing entries in a record: {sorted(missing_keys)}')
            return reformatted
        else:
            raise ValueError(f'expected a dict at {".".join(context.path)}')

    def serialize_variant(self, context: TypeEvaluationContext, tt: VariantType, obj: Any) -> Any:
        from collections import Mapping
        if isinstance(obj, Mapping):
            proposed_variants = {}
            error_variants = {}
            reformatted = unflatten_dotted_keys(obj)
            for name, field_type in tt.named_args:
                if name in reformatted:
                    try:
                        new_value = self._serialize_dispatch(
                            context.append_path(name), field_type, reformatted[name])
                        proposed_variants[name] = new_value
                    except:
                        error_variants[name] = reformatted[name]

            if not proposed_variants:
                raise ValueError('got an empty dict where a single-key dict was expected')
            if len(proposed_variants) == 1:
                return proposed_variants
            else:
                non_optional_ctors = dict()
                optional_ctors = set()
                for name, field_type in tt.named_args:
                    if name in proposed_variants:
                        value = proposed_variants[name]
                        if can_be_optional(field_type, value):
                            optional_ctors.add(name)
                        else:
                            non_optional_ctors[name] = value
                if len(non_optional_ctors) == 1:
                    return non_optional_ctors
                else:
                    raise ValueError(f'too many entries in a variant dict: could not choose among '
                                     f'{sorted(non_optional_ctors)}')
        else:
            error = 'variants must be encoded as dictionaries with a single key and a value'
            raise ValueError(f'{error} (got {obj!r} instead)', obj)

    def serialize_unsupported(self, context: TypeEvaluationContext, tt: UnsupportedType, obj: Any) \
            -> Any:
        return obj


def can_be_optional(tt: Type, obj: Any):
    from collections import Mapping
    if not obj and obj != 0:
        return True
    if isinstance(obj, Mapping):
        # TODO: For now we're not using type information; might be useful to introduce it at some
        #  point though
        cbo = all(can_be_optional(None, value) for value in obj.values())
        return cbo
    else:
        return False




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from aiohttp import ClientSession
from aiohttp.web import get, post, Request, Response
from asyncio import get_event_loop
from oauthlib.oauth2 import WebApplicationClient
from typing import Dict
import uuid

from ..model.network import OAuthSettings


class OAuthHandler:

    AUTH_ROUTE = '/auth/callback'

    def __init__(self, url_prefix: str):
        self.url_prefix = url_prefix
        self.callbacks = {}  # type: Dict[str, OAuthAuthorizationCodeGrantRequest]
        self.routes = [
            get(self.AUTH_ROUTE, self.oauth_callback)
        ]

    async def auth_flow(self, settings: OAuthSettings) -> OAuthSettings:
        """
        Perform an authorization flow with the remote server and prepare the server for handling
        a response.

        :return:
            :class:`OAuthSettings` populated with an initial token and possibly information for
            refreshing that token.
        """
        oauth_req = OAuthAuthorizationCodeGrantRequest(self.url_prefix + self.AUTH_ROUTE, settings)
        self.callbacks[oauth_req.state] = oauth_req
        await oauth_req.start()

        # wait until the callback has been called
        return await oauth_req.future

    async def oauth_callback(self, request: Request) -> Response:
        code = request.query.get('code')
        state = request.query.get('state')
        oauth_req = self.callbacks.pop(state)
        if oauth_req is not None:
            try:
                await oauth_req.handle_redirect(str(request.url), code, state)
                return await self._auth_finished(request)
            finally:
                await oauth_req.close()
        else:
            return await self._invalid_state(request)

    async def _auth_finished(self, request: Request) -> Response:
        pass

    async def _invalid_state(self, request: Request) -> Response:
        pass


class OAuthAuthorizationCodeGrantRequest:
    def __init__(self, redirect_uri: str, settings: OAuthSettings):
        self.settings = settings
        self.future = get_event_loop().create_future()
        self.state = uuid.uuid4().hex
        self.client = WebApplicationClient(settings.client_id)
        self.redirect_uri = redirect_uri
        self.session = ClientSession()

    async def start(self) -> None:
        uri = self.client.prepare_request_uri(
            self.settings.token_uri, redirect_uri=self.redirect_uri, state=self.state)
        body = self.client.prepare_request_body()

        async with self.session.post(uri, data=body) as response:
            print(response.status)
            print(await response.text())

    async def handle_redirect(self, uri: str, code: str, state: str) -> OAuthSettings:
        self.client.parse_request_uri_response(uri, state)

        kwargs = {}
        if self.settings.client_secret:
            kwargs['client_secret'] = self.settings.client_secret

        uri = self.settings.token_uri
        body = self.client.prepare_request_body(
            code=code, redirect_uri=self.redirect_uri, **kwargs)

        # fetch token

        self.client.parse_request_body_response(response_body)

        new_settings = OAuthSettings(

        )

        self.future.set_result(new_settings)
        return new_settings

    async def close(self):
        await self.session.close()
